#
#  Copyright (C) 2016 Canonical Ltd.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import clickreviews.remote

USER_DATA_FILE = os.path.join(clickreviews.remote.DATA_DIR,
                              'snapd-base-declaration.yaml')

BD_DATA_URL = \
    ("https://raw.githubusercontent.com/ubports/click-reviewers-tools/"
     "xenial/data/snapd-base-declaration.yaml")


def get_base_declaration_file(fn):
    if fn is None:
        fn = USER_DATA_FILE
    clickreviews.remote.get_remote_file(fn, BD_DATA_URL)


class SnapdBaseDeclaration(object):
    def __init__(self, local_copy_fn=None):
        self.decl = clickreviews.remote.read_cr_file(USER_DATA_FILE,
                                                     BD_DATA_URL,
                                                     local_copy_fn,
                                                     as_yaml=True)



'''sr_common.py: common classes and functions'''
#
# Copyright (C) 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import os
import re
import yaml


from clickreviews.common import (
    Review,
    ReviewException,
    error,
    open_file_read,
)

import clickreviews.snapd_base_declaration as snapd_base_declaration


#
# Utility classes
#
class SnapReviewException(ReviewException):
    '''This class represents SnapReview exceptions'''


class SnapReview(Review):
    '''This class represents snap reviews'''
    snappy_required = ["name",
                       "version",
                       ]
    # optional snappy fields here (may be required by appstore)
    snappy_optional = ['apps',
                       'assumes',
                       'architectures',
                       'confinement',
                       'description',
                       'environment',
                       'epoch',
                       'grade',
                       'hooks',
                       'icon',
                       'license-agreement',  # OBSOLETE (LP: #1638303)
                       'license-version',    # OBSOLETE
                       'summary',
                       'type',
                       'plugs',
                       'slots',
                       ]

    apps_required = ['command']
    apps_optional = ['aliases',
                     'completer',
                     'daemon',
                     'environment',
                     'stop-command',
                     'stop-timeout',
                     'restart-condition',
                     'post-stop-command',
                     'plugs',
                     'slots',
                     'ports',
                     'socket',
                     'listen-stream',
                     'socket-user',
                     'socket-group',
                     ]
    hooks_required = []
    hooks_optional = ['plugs']

    # Valid values for 'type' in packaging yaml
    # - app
    # - core
    # - kernel
    # - gadget
    # - os (deprecated)
    valid_snap_types = ['app',
                        'core',
                        'kernel',
                        'gadget',
                        'os',
                        ]

    # https://snapcraft.io/docs/interface-management
    # 'plugs':
    #    'interface': name
    #    'attrib-name': <type>
    # 'slots':
    #    'interface': name
    #    'attrib-name': <type>
    # self.interfaces lists interfaces and the valid attribute names for the
    # interface with the valid python type for the attribute (eg, [], '', {},
    # etc).  # Interfaces with no attributes should specify an empty
    # dictionary.
    #
    # Interfaces are read from the base declaration in __init__() so they don't
    # have to be added to self.interfaces.
    interfaces = dict()

    # interfaces_attribs[iface] contains all known attributes and will be
    # merged into self.interfaces after reading the base declaration since
    # the base declaration doesn't declare all the known attributes.
    interfaces_attribs = {'bool-file': {'path/slots': ""},
                          'browser-support': {'allow-sandbox/plugs': False},
                          'content': {'read/slots': [],
                                      'write/slots': [],
                                      'target/plugs': "",
                                      'default-provider/plugs': "",
                                      'content/plugs': "",
                                      'content/slots': "",
                                      },
                          'dbus': {'name/slots': "",
                                   'bus/slots': "",
                                   'name/plugs': "",
                                   'bus/plugs': "",
                                   },
                          'docker-support': {'privileged-containers/plugs':
                                             False},
                          'gpio': {'number/slots': 0},
                          'hidraw': {'path/slots': "",
                                     'usb-vendor/slots': 0,
                                     'usb-product/slots': 0,
                                     },
                          'i2c': {'path/slots': ""},
                          'iio': {'path/slots': ""},
                          'mpris': {'name/slots': ""},
                          'serial-port': {'path/slots': "",
                                          'usb-vendor/slots': 0,
                                          'usb-product/slots': 0,
                                          },
                          }

    # interfaces_required[iface] lists required attributes as combinations.
    # Eg, ['a', 'b', 'c/d'] means one of 'a', 'b', or 'c and d' is required.
    # This is to avoid situations like:
    # https://forum.snapcraft.io/t/broken-snap-breaking-snapd/401/8
    interfaces_required = {
        'bool-file': {'slots': ['path']},
        'content': {
            'slots': ['read', 'write'],
            'plugs': ['target'],
        },
        'dbus': {
            'slots': ['name/bus'],
            'plugs': ['name/bus'],
        },
        'gpio': {'slots': ['number']},
        'hidraw': {
            'slots': [
                'path',
                'path/!usb-vendor/!usb-product',
                'path/usb-vendor/usb-product'
            ],
        },
        'i2c': {'slots': ['path']},
        'iio': {'slots': ['path']},
        'serial-port': {
            'slots': [
                'path/!usb-vendor/!usb-product',
                'path/usb-vendor/usb-product'
            ],
        },
    }

    # In progress interfaces are those that are not yet in snapd but for
    # some reason we need them. Normally we will never want to do this, but
    # for example the unity8 interface wass in progress and they wanted CI
    # uploads
    inprogress_interfaces = {
        '16': {
            'plugs': {},
            'slots': {},
        }
    }

    def __init__(self, fn, review_type, overrides=None):
        Review.__init__(self, fn, review_type, overrides=overrides)

        if not self.is_snap2:
            return

        snap_yaml = self._extract_snap_yaml()
        try:
            self.snap_yaml = yaml.safe_load(snap_yaml)
        except Exception:  # pragma: nocover
            error("Could not load snap.yaml. Is it properly formatted?")

        # If local_copy is None, then this will check the server to see if
        # we are up to date. However, if we are working within the development
        # tree, use it unconditionally.
        local_copy = None
        branch_fn = os.path.join(os.path.dirname(__file__),
                                 '../data/snapd-base-declaration.yaml')
        if os.path.exists(branch_fn):
            local_copy = branch_fn
        p = snapd_base_declaration.SnapdBaseDeclaration(local_copy)
        # FIXME: don't hardcode series
        self.base_declaration_series = "16"
        self.base_declaration = p.decl[self.base_declaration_series]

        # Add in-progress interfaces
        if self.base_declaration_series in self.inprogress_interfaces:
            rel = self.base_declaration_series
            for side in ['plugs', 'slots']:
                if side not in self.base_declaration or \
                        side not in self.inprogress_interfaces[rel]:
                    continue

                if side == 'plugs':
                    oside = 'slots'
                else:
                    oside = 'plugs'

                for iface in self.inprogress_interfaces[rel][side]:
                    if iface in self.base_declaration[side] or \
                            iface in self.base_declaration[oside]:
                        # don't override anything in the base declaration
                        continue
                    self.base_declaration[side][iface] = \
                        self.inprogress_interfaces[rel][side][iface]

        # to simplify checks, gather up all the interfaces into one dict()
        for side in ['plugs', 'slots']:
            for k in self.base_declaration[side]:
                if k in self.interfaces_attribs:
                    self.interfaces[k] = self.interfaces_attribs[k]
                else:
                    self.interfaces[k] = {}

        # default to 'app'
        if 'type' not in self.snap_yaml:
            self.snap_yaml['type'] = 'app'

        if 'architectures' in self.snap_yaml:
            self.pkg_arch = self.snap_yaml['architectures']
        else:
            self.pkg_arch = ['all']

        self.is_snap_gadget = False
        if 'type' in self.snap_yaml and self.snap_yaml['type'] == 'gadget':
            self.is_snap_gadget = True

        # snapd understands:
        #   plugs:
        #     foo: null
        # but yaml.safe_load() treats 'null' as 'None', but we need a {}, so
        # we need to account for that.
        for k in ['plugs', 'slots']:
            if k not in self.snap_yaml:
                continue
            for iface in self.snap_yaml[k]:
                if not isinstance(self.snap_yaml[k], dict):
                    # eg, top-level "plugs: [ content ]"
                    error("Invalid top-level '%s'"
                          " (not a dict)" % k)  # pragma: nocover
                if self.snap_yaml[k][iface] is None:
                    self.snap_yaml[k][iface] = {}

    # Since coverage is looked at via the testsuite and the testsuite mocks
    # this out, don't cover this
    def _extract_snap_yaml(self):  # pragma: nocover
        '''Extract and read the snappy 16.04 snap.yaml'''
        y = os.path.join(self.unpack_dir, "meta/snap.yaml")
        if not os.path.isfile(y):
            error("Could not find snap.yaml.")
        return open_file_read(y)

    # Since coverage is looked at via the testsuite and the testsuite mocks
    # this out, don't cover this
    def _get_unpack_dir(self):  # pragma: nocover
        '''Get unpack directory'''
        return self.unpack_dir

    def _verify_pkgname(self, n):
        '''Verify package name'''
        # From validSnapName in snapd/snap/validate.go:
        #   "^(?:[a-z0-9]+-?)*[a-z](?:-?[a-z0-9])*$"
        # but this regex is very inefficient and certain names will make python
        # work extremely hard. Instead we use this and make sure the name isn't
        # all digits.
        pat = re.compile(r'^[a-z0-9](?:-?[a-z0-9])*$')

        if pat.search(n) and not n.isnumeric():
            return True
        return False

    def _verify_appname(self, n):
        '''Verify app name'''
        pat = re.compile(r'^[a-zA-Z0-9](?:-?[a-zA-Z0-9])*$')

        if pat.search(n):
            return True
        return False



'''sr_security.py: snap security checks'''
#
# Copyright (C) 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from clickreviews.sr_common import (
    SnapReview,
)
from clickreviews.common import (
    cmd,
    create_tempdir,
    ReviewException,
    AA_PROFILE_NAME_MAXLEN,
    AA_PROFILE_NAME_ADVLEN,
    MKSQUASHFS_OPTS,
)
from clickreviews.overrides import (
    sec_mode_overrides,
    sec_browser_support_overrides,
)
import os
import re


class SnapReviewSecurity(SnapReview):
    '''This class represents snap security reviews'''
    def __init__(self, fn, overrides=None):
        SnapReview.__init__(self, fn, "security-snap-v2", overrides=overrides)

        if not self.is_snap2:
            return

        self.sec_skipped_types = ['oem',
                                  'os',
                                  'kernel']  # these don't need security items

    def _unsquashfs_lls(self, snap_pkg):
        '''Run unsquashfs -lls on a snap package'''
        return cmd(['unsquashfs', '-lls', snap_pkg])

    def check_security_plugs_browser_support_with_daemon(self):
        '''Check security plugs - browser-support not used with daemon'''
        def _plugref_is_interface(ref, iface):
            if ref == iface:
                return True
            elif 'plugs' in self.snap_yaml and \
                    ref in self.snap_yaml['plugs'] and \
                    'interface' in self.snap_yaml['plugs'][ref] and \
                    self.snap_yaml['plugs'][ref]['interface'] == iface:
                return True
            return False

        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        found_app_plugs = False
        for app in self.snap_yaml['apps']:
            if 'plugs' in self.snap_yaml['apps'][app]:
                found_app_plugs = True
                break

        if not found_app_plugs and 'plugs' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            if found_app_plugs and 'plugs' not in self.snap_yaml['apps'][app]:
                continue
            elif 'plugs' in self.snap_yaml['apps'][app]:
                plugs = self.snap_yaml['apps'][app]['plugs']
            else:
                plugs = self.snap_yaml['plugs']

            if 'daemon' not in self.snap_yaml['apps'][app]:
                continue

            for plug_ref in plugs:
                if _plugref_is_interface(plug_ref, "browser-support"):
                    if self.snap_yaml['name'] in \
                            sec_browser_support_overrides:
                        t = 'info'
                        s = "OK (allowing 'daemon' with 'browser-support'"
                    else:
                        t = 'warn'
                        s = "(NEEDS REVIEW) 'daemon' should not be used " + \
                            "with 'browser-support'"
                    n = self._get_check_name('daemon_with_browser-support',
                                             app=app)
                    self._add_result(t, n, s, manual_review=True)

    def check_apparmor_profile_name_length(self):
        '''Check AppArmor profile name length'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        maxlen = AA_PROFILE_NAME_MAXLEN
        advlen = AA_PROFILE_NAME_ADVLEN

        for app in self.snap_yaml['apps']:
            t = 'info'
            n = self._get_check_name('profile_name_length', app=app)
            s = "OK"
            profile = "snap.%s.%s" % (self.snap_yaml['name'], app)
            if len(profile) > maxlen:
                t = 'error'
                s = ("'%s' too long (exceeds %d characters). Please shorten "
                     "'%s' and/or '%s'" % (profile, maxlen,
                                           self.snap_yaml['name'], app))
            elif len(profile) > advlen:
                t = 'warn'
                s = ("'%s' is long (exceeds %d characters) and thus could be "
                     "problematic in certain environments. Please consider "
                     "shortening '%s' and/or '%s'" % (profile, advlen,
                                                      self.snap_yaml['name'],
                                                      app))
            self._add_result(t, n, s)

    def check_squashfs_resquash(self):
        '''Check resquash of squashfs'''
        if not self.is_snap2:
            return

        fn = os.path.abspath(self.pkg_filename)

        # Verify squashfs supports the -fstime option, if not, warn (which
        # blocks in store)
        (rc, out) = cmd(['unsquashfs', '-fstime', fn])
        if rc != 0:
            t = 'warn'
            n = self._get_check_name('squashfs_supports_fstime')
            s = 'could not determine fstime of squashfs'
            self._add_result(t, n, s)
            return
        fstime = out.strip()

        # For now, skip the checks on if have symlinks due to LP: #1555305
        (rc, out) = cmd(['unsquashfs', '-lls', fn])
        if rc != 0:
            t = 'error'
            n = self._get_check_name('squashfs_lls')
            s = 'could not list contents of squashfs'
            self._add_result(t, n, s)
            return
        elif 'lrwxrwxrwx' in out:
            t = 'info'
            n = self._get_check_name('squashfs_resquash_1555305')
            s = 'cannot reproduce squashfs'
            link = 'https://launchpad.net/bugs/1555305'
            self._add_result(t, n, s, link=link)
            return
        # end LP: #1555305 workaround

        tmpdir = create_tempdir()  # this is autocleaned
        tmp_unpack = os.path.join(tmpdir, 'squashfs-root')
        tmp_repack = os.path.join(tmpdir, 'repack.snap')

        curdir = os.getcwd()
        os.chdir(tmpdir)
        # ensure we don't alter the permissions from the unsquashfs
        old_umask = os.umask(000)

        try:
            (rc, out) = cmd(['unsquashfs', '-d', tmp_unpack, fn])
            if rc != 0:
                raise ReviewException("could not unsquash '%s': %s" %
                                      (os.path.basename(fn), out))
            (rc, out) = cmd(['mksquashfs', tmp_unpack, tmp_repack,
                             '-fstime', fstime] + MKSQUASHFS_OPTS)
            if rc != 0:
                raise ReviewException("could not mksquashfs '%s': %s" %
                                      (os.path.relpath(tmp_unpack, tmpdir),
                                       out))
        except ReviewException as e:
            t = 'error'
            n = self._get_check_name('squashfs_resquash')
            self._add_result(t, n, str(e))
            return
        finally:
            os.umask(old_umask)
            os.chdir(curdir)

        # Now calculate the hashes
        t = 'info'
        n = self._get_check_name('squashfs_repack_checksum')
        s = "OK"

        (rc, out) = cmd(['sha512sum', fn])
        if rc != 0:
            t = 'error'
            s = "could not determine checksum of '%s'" % os.path.basename(fn)
            self._add_result(t, n, s)
            return
        orig_sum = out.split()[0]

        (rc, out) = cmd(['sha512sum', tmp_repack])
        if rc != 0:
            t = 'error'
            s = "could not determine checksum of '%s'" % \
                os.path.relpath(tmp_repack, tmpdir)
            self._add_result(t, n, s)
            return
        repack_sum = out.split()[0]

        if orig_sum != repack_sum:
            if 'type' in self.snap_yaml and self.snap_yaml['type'] == 'os':
                t = 'info'
                s = 'checksums do not match (expected for os snap)'
            else:
                # FIXME: turn this into an error once the squashfs-tools bugs
                # are fixed
                # t = 'error'
                t = 'info'
                s = "checksums do not match. Please ensure the snap is " + \
                    "created with either 'snapcraft snap <DIR>' or " + \
                    "'mksquashfs <dir> <snap> %s'" % " ".join(MKSQUASHFS_OPTS)
        self._add_result(t, n, s)

    def check_squashfs_files(self):
        '''Check squashfs files'''
        def _check_allowed_perms(mode, allowed):
            for p in mode[1:]:
                if p not in allowed:
                    return False
            return True

        if not self.is_snap2:
            return

        pkgname = self.snap_yaml['name']

        snap_type = 'app'
        if 'type' in self.snap_yaml:
            snap_type = self.snap_yaml['type']

        fn = os.path.abspath(self.pkg_filename)

        (rc, out) = self._unsquashfs_lls(fn)
        if rc != 0:
            t = 'error'
            n = self._get_check_name('squashfs_files_unsquash')
            s = 'unsquashfs -lls <snap> failed'
            self._add_result(t, n, s)
            return

        in_header = True
        malformed = []
        errors = []

        fname_pat = re.compile(r'.* squashfs-root')
        date_pat = re.compile(r'^\d\d\d\d-\d\d-\d\d$')
        time_pat = re.compile(r'^\d\d:\d\d$')
        mknod_pat_full = re.compile(r'.,.')
        count = 0

        for line in out.splitlines():
            count += 1
            if in_header:
                if len(line) < 1:
                    in_header = False
                continue

            tmp = line.split()
            if len(tmp) < 6:
                malformed.append("wrong number of fields in '%s'" % line)
                continue

            fname = fname_pat.sub('.', line)
            ftype = tmp[0][0]

            # Also see 'info ls', but we list only the Linux ones
            ftype = line[0]
            if ftype not in ['b', 'c', 'd', 'l', 'p', 's', '-']:
                errors.append("unknown type '%s' for entry '%s'" % (ftype,
                                                                    fname))
                continue

            # verify mode
            mode = tmp[0][1:]
            if len(mode) != 9:
                malformed.append("mode '%s' malformed for '%s'" % (mode,
                                                                   fname))
                continue
            if ftype == 'd' or ftype == '-':
                perms = ['r', 'w', 'x', '-']
                if ftype == 'd':  # allow sticky directories for stage-packages
                    perms.append('t')
                if not _check_allowed_perms(mode, perms):
                    if pkgname not in sec_mode_overrides or \
                        fname not in sec_mode_overrides[pkgname] or \
                            sec_mode_overrides[pkgname][fname] != mode:
                        errors.append("unusual mode '%s' for entry '%s'" %
                                      (mode, fname))
                        continue
                # No point checking for world-writable, the squashfs is
                # readonly
                # if mode[-2] != '-':
                #     errors.append("'%s' is world-writable" % fn)
                #     continue
            elif ftype == 'l':
                if mode != 'rwxrwxrwx':
                    errors.append("unusual mode '%s' for symlink '%s'" %
                                  (mode, fname))
                    continue
            else:
                if snap_type != 'os':
                    errors.append("file type '%s' not allowed (%s)" % (ftype,
                                                                       fname))
                    continue

            # verify user and group
            if '/' not in tmp[1]:
                malformed.append("user/group '%s' malformed for '%s'" %
                                 (tmp[1], fname))
                continue
            (user, group) = tmp[1].split('/')
            # we enforce 'root/root'
            if snap_type != 'os' and (user != 'root' or group != 'root'):
                errors.append("unusual user/group '%s' for '%s'" % (tmp[1],
                                                                    fname))
                continue

            date_idx = 3
            time_idx = 4
            if ftype == 'b' or ftype == 'c':
                # Account for unsquashfs -lls doing:
                # crw-rw-rw- root/root             1,  8 2016-08-09 ...
                # crw-rw---- root/root            10,141 2016-08-09 ...

                if mknod_pat_full.search(tmp[2]):
                    (major, minor) = tmp[2].split(',')
                else:
                    date_idx = 4
                    time_idx = 5
                    major = tmp[2][:-1]
                    minor = tmp[3]

                try:
                    int(major)
                except ValueError:
                    malformed.append("major '%s' malformed for '%s'" %
                                     (major, fname))
                try:
                    int(minor)
                except ValueError:
                    malformed.append("minor '%s' malformed for '%s'" %
                                     (minor, fname))
            else:
                size = tmp[2]
                try:
                    int(size)
                except ValueError:
                    malformed.append("size '%s' malformed for '%s'" % (size,
                                                                       fname))
                    continue

            date = tmp[date_idx]
            if not date_pat.search(date):
                malformed.append("date '%s' malformed for '%s'" % (date,
                                                                   fname))
                continue

            time = tmp[time_idx]
            if not time_pat.search(time):
                malformed.append("time '%s' malformed for '%s'" % (time,
                                                                   fname))
                continue

        if count < 4:
            t = 'error'
            n = self._get_check_name('squashfs_files_malformed output')
            s = "unsquashfs -lls ouput too short"
            self._add_result(t, n, s)

        if len(malformed) > 0:
            t = 'error'
            n = self._get_check_name('squashfs_files_malformed_line')
            s = "malformed lines in unsquashfs output: '%s'" % \
                ", ".join(malformed)
            self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('squashfs_files')
        s = 'OK'
        if len(errors) > 0:
            t = 'error'
            s = "found errors in file output: %s" % ", ".join(errors)
        self._add_result(t, n, s)



'''sr_declaration.py: click declaration'''
#
# Copyright (C) 2014-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
from clickreviews.sr_common import SnapReview, SnapReviewException
from clickreviews.overrides import iface_attributes_noflag
import re


class SnapDeclarationException(SnapReviewException):
    '''This class represents SnapDeclaration exceptions'''


class SnapReviewDeclaration(SnapReview):
    '''This class represents click lint reviews'''
    def __init__(self, fn, overrides=None):
        SnapReview.__init__(self, fn, "declaration-snap-v2",
                            overrides=overrides)

        if not self.is_snap2:
            return

        self._verify_declaration(self.base_declaration, base=True)

        self.snap_declaration = None
        if overrides is not None and ('snap_decl_plugs' in overrides or
                                      'snap_decl_slots' in overrides):
            self.snap_declaration = {}
            self.snap_declaration = {'plugs': {}, 'slots': {}}
            if 'snap_decl_plugs' in overrides:
                self.snap_declaration['plugs'] = overrides['snap_decl_plugs']
            if 'snap_decl_slots' in overrides:
                self.snap_declaration['slots'] = overrides['snap_decl_slots']

            self._verify_declaration(self.snap_declaration, base=False)

    def is_bool(self, item):
        if isinstance(item, int) and (item is True or item is False):
            return True
        return False

    def str2bool(self, s):
        if s == "true" or s == "True":
            return True
        if s == "false" or s == "False":
            return False
        return s

    def _verify_declaration(self, decl, base=False):
        '''Verify declaration'''
        def malformed(name, s, base=False):
            pre = ""
            if base:
                pre = "base "
            err = "%sdeclaration malformed (%s)" % (pre, s)
            if base:
                raise SnapDeclarationException(err)
            self._add_result('error', name, err)

        def verify_constraint(cstr, decl, key, iface, index, allowed,
                              has_alternates):
            found_errors = False
            if self.is_bool(cstr):
                if not base:
                    self._add_result('info', n, s)
                return True
            elif not isinstance(cstr, dict):
                malformed(n, "%s not True, False or dict" %
                          constraint, base)
                return True

            for cstr_key in cstr:
                if cstr_key not in allowed:
                    name = self._get_check_name('valid_%s' % key, app=iface,
                                                extra="%s_%s" % (constraint,
                                                                 cstr_key))
                    malformed(name, "unknown constraint key '%s'" % cstr_key,
                              base)
                    found_errors = True

            cstr_bools = ["on-classic"]
            cstr_lists = ["plug-snap-type",
                          "slot-snap-type",
                          "plug-publisher-id"
                          "slot-publisher-id",
                          "plug-snap-id",
                          "slot-snap-id"
                          ]
            cstr_dicts = ["plug-attributes", "slot-attributes"]
            for cstr_key in cstr:
                badn = self._get_check_name('valid_%s' % key, app=iface,
                                            extra="%s_%s" % (constraint,
                                                             cstr_key))
                if cstr_key in cstr_bools:
                    # snap declarations from the store express bools as
                    # strings
                    if isinstance(cstr[cstr_key], str):
                        cstr[cstr_key] = \
                            self.str2bool(cstr[cstr_key])
                        if has_alternates:
                            decl[key][iface][constraint][index][cstr_key] = \
                                self.str2bool(decl[key][iface][constraint][index][cstr_key])  # noqa
                    if not self.is_bool(cstr[cstr_key]):
                        malformed(badn, "'%s' not True or False" % cstr_key,
                                  base)
                        found_errors = True
                elif cstr_key in cstr_lists:
                    if not isinstance(cstr[cstr_key], list):
                        malformed(badn, "'%s' not a list" % cstr_key, base)
                        found_errors = True
                    else:
                        for entry in cstr[cstr_key]:
                            if not isinstance(entry, str):
                                malformed(badn, "'%s' in '%s' not a string" %
                                          (entry, cstr_key), base)
                                found_errors = True
                elif cstr_key in cstr_dicts:
                    if not isinstance(cstr[cstr_key], dict):
                        malformed(badn, "'%s' not a dict" % cstr_key, base)
                        found_errors = True
                    else:
                        for attrib in cstr[cstr_key]:
                            bn = self._get_check_name('valid_%s' % key,
                                                      app=iface,
                                                      extra="%s_%s" %
                                                      (constraint, cstr_key))
                            if iface not in self.interfaces_attribs:
                                malformed(bn, "unknown attribute '%s'" %
                                          attrib,
                                          base)
                                found_errors = True
                                continue

                            found_iface_attr = False
                            for tmp in self.interfaces_attribs[iface]:
                                known, side = tmp.split('/')
                                if attrib != known:
                                    continue
                                spec_side = side[:-1]

                                if cstr_key.startswith(spec_side):
                                    found_iface_attr = True

                                # snap declarations from the store express
                                # bools as strings
                                if isinstance(cstr[cstr_key][attrib], str):
                                    cstr[cstr_key][attrib] = \
                                        self.str2bool(cstr[cstr_key][attrib])
                                    if has_alternates:
                                        decl[key][iface][constraint][index][cstr_key][attrib] = self.str2bool(decl[key][iface][constraint][index][cstr_key][attrib])  # noqa

                                attr_type = cstr[cstr_key][attrib]
                                tmp_attr = self.interfaces_attribs[iface][tmp]

                                if not isinstance(attr_type, type(tmp_attr)):
                                    malformed(bn,
                                              "wrong type '%s' for attribute "
                                              "'%s'"
                                              % (attr_type, attrib),
                                              base)
                                    found_errors = True
                                    break

                            if not found_iface_attr:
                                malformed(bn,
                                          "attribute '%s' wrong for '%ss'" %
                                          (attrib, cstr_key[:4]),
                                          base)
                                found_errors = True

                if not found_errors and \
                        cstr_key == "plug-publisher-id" or \
                        cstr_key == "slot-publisher-id":
                    for pubid in cstr[cstr_key]:
                        if not pub_pat.search(pubid):
                            malformed(n, "invalid format for publisher id '%s'"
                                      % pubid)
                            found_errors = True
                            break
                        if pubid.startswith('$'):
                            if cstr_key == "plug-publisher-id" and \
                                    pubid != "$SLOT_PUBLISHER_ID":
                                malformed(n,
                                          "invalid publisher id '%s'" % pubid)
                                found_errors = True
                                break
                            elif cstr_key == "slot-publisher-id" and \
                                    pubid != "$PLUG_PUBLISHER_ID":
                                malformed(n,
                                          "invalid publisher id '%s'" % pubid)
                                found_errors = True
                                break
                elif not found_errors and \
                        cstr_key == "plug-snap-id" or \
                        cstr_key == "slot-snap-id":
                    for id in cstr[cstr_key]:
                        if not id_pat.search(id):
                            malformed(n,
                                      "invalid format for snap id '%s'" % id)
                            found_errors = True
                            break
                elif not found_errors and \
                        cstr_key == "plug-snap-type" or \
                        cstr_key == "slot-snap-type":
                    for snap_type in cstr[cstr_key]:
                        if snap_type not in self.valid_snap_types:
                            malformed(n, "invalid snap type '%s'" % snap_type)
                            found_errors = True
                            break

            return found_errors
            # end verify_constraint()

        # from snapd.git/assers/ifacedecls.go
        id_pat = re.compile(r'^[a-z0-9A-Z]{32}$')
        pub_pat = re.compile(r'^(?:[a-z0-9A-Z]{32}|[-a-z0-9]{2,28}|'
                             r'\$[A-Z][A-Z0-9_]*)$')

        if not isinstance(decl, dict):
            malformed(self._get_check_name('valid_dict'), "not a dict", base)
            return
        elif len(decl) == 0:
            malformed(self._get_check_name('valid_dict'), "empty", base)
            return

        for key in decl:
            if key not in ["plugs", "slots"]:
                malformed(self._get_check_name('valid_key'),
                          "unknown key '%s'" % key, base)
                return

            if not isinstance(decl[key], dict):
                malformed(self._get_check_name('valid_dict', app=key),
                          "not a dict", base)
                return

            for iface in decl[key]:
                # snap declarations from the store express bools as strings
                if isinstance(decl[key][iface], str):
                    decl[key][iface] = self.str2bool(decl[key][iface])
                # iface may be bool or dict
                if self.is_bool(decl[key][iface]):
                    n = self._get_check_name('valid_%s' % key, app=iface)
                    self._add_result('info', n, 'OK')
                    continue
                elif not isinstance(decl[key][iface], dict):
                    malformed(self._get_check_name('valid_%s_dict' % key,
                                                   app=iface),
                              "interface not True, False or dict", base)
                    continue

                found_errors = False
                for constraint in decl[key][iface]:
                    # snap declarations from the store express bools as strings
                    if isinstance(decl[key][iface][constraint], str):
                        decl[key][iface][constraint] = \
                            self.str2bool(decl[key][iface][constraint])

                    t = 'info'
                    n = self._get_check_name('valid_%s' % key, app=iface,
                                             extra=constraint)
                    s = "OK"
                    cstr = decl[key][iface][constraint]

                    allowed_ctrs = ["allow-installation",
                                    "deny-installation",
                                    "allow-connection",
                                    "allow-auto-connection",
                                    "deny-connection",
                                    "deny-auto-connection"
                                    ]
                    if constraint not in allowed_ctrs:
                        malformed(n, "unknown constraint '%s'" % constraint,
                                  base)
                        break

                    allowed = []
                    if constraint.endswith("-installation"):
                        allowed = ["on-classic"]
                        if key == "plugs":
                            allowed.append("plug-snap-type")
                            allowed.append("plug-attributes")
                        elif key == "slots":
                            allowed.append("slot-snap-type")
                            allowed.append("slot-attributes")
                    else:
                        allowed = ["plug-attributes", "slot-attributes",
                                   "on-classic"]
                        if key == "plugs":
                            allowed.append("slot-publisher-id")
                            allowed.append("slot-snap-id")
                            allowed.append("slot-snap-type")
                        elif key == "slots":
                            allowed.append("plug-publisher-id")
                            allowed.append("plug-snap-id")
                            allowed.append("plug-snap-type")

                    # constraint may be bool or dict or lists of bools and
                    # dicts
                    alternates = []
                    if isinstance(cstr, list):
                        alternates = cstr
                    else:
                        alternates.append(cstr)

                    index = 0
                    for alt in alternates:
                        if verify_constraint(alt, decl, key, iface, index,
                                             allowed, (len(alternates) > 1)):
                            found_errors = True
                        index += 1

                    if not base and not found_errors:
                        self._add_result(t, n, s)

    def _match(self, against, val):
        '''Ordering matters since 'against' is treated as a regex if str'''
        if type(against) != type(val):
            return False

        if type(val) not in [str, list, dict, bool]:
            raise SnapDeclarationException("unknown type '%s'" % val)

        matched = False

        if isinstance(val, str):
            if re.search(r'^(%s)$' % against, val):
                matched = True
        elif isinstance(val, list):
            matched = (sorted(against) == sorted(val))
        else:  # bools and dicts (TODO: nested matches for dicts)
            matched = (against == val)

        return matched

    def _search(self, d, key, val=None, subkey=None, subval=None,
                subval_inverted=False):
        '''Search dictionary 'd' for matching values. Returns true when
           - val == d[key]
           - subval in d[key][subkey]
           - subval dictionary has any matches in d[key][subkey] dict
           - subval_inverted == True and subval has any non-matches in
             d[key][subkey] dict

           When 'key' must be in 'd' and when 'subkey' is not None, it must be
           in d[key] (we want to raise an Exception here for when _search() is
           used with only exact matches).
        '''
        found = False

        if val is not None and val == d[key]:
            found = True
        elif isinstance(d[key], dict) and subkey is not None and \
                subval is not None:
            if self.is_bool(d[key][subkey]):
                found = d[key][subkey] == subval
                if subval_inverted:
                    found = not found
            elif isinstance(d[key][subkey], list):
                if subval_inverted:
                    if subval not in d[key][subkey]:
                        found = True
                elif subval in d[key][subkey]:
                    found = True
            elif isinstance(d[key][subkey], dict) and isinstance(subval, dict):
                d_keys = set(d[key][subkey].keys())
                subval_keys = set(subval.keys())
                int_keys = d_keys.intersection(subval_keys)
                matches = 0
                for subsubkey in int_keys:
                    if self._match(d[key][subkey][subsubkey],
                                   subval[subsubkey]) or \
                            d[key][subkey][subsubkey] in \
                            iface_attributes_noflag:
                        found = True
                        matches += 1

                if subval_inverted:
                    # return true when something didn't match
                    if matches != len(int_keys):
                        found = True
                    else:
                        found = False

        return found

    def _get_decl(self, base, snap, side, interface, dtype):
        '''If the snap declaration has something to say about the declaration
           override type (dtype), then use it instead of the base declaration.
        '''
        decl = base
        base_decl = True
        decl_type = "base"

        if snap is not None and side in snap and interface in snap[side]:
            for k in snap[side][interface]:
                if k.endswith(dtype):
                    decl = snap
                    base_decl = False
                    decl_type = "snap"
                    break

        return (decl, base_decl, decl_type)

    def _get_all_combinations(self, interface):
        '''Return list of all base and snap declaration combinations where
           each base/snap declaration pair represents a particular combination
           of alternate constraints. Also return if there are alternate
           constraints anywhere.

           For simple declarations, this will return the interface of the
           base declaration and if a snap declaration is specified, the
           interface of the snap declaration (ie, a single base/snap
           declaration pair).

           For complex declarations with alternate constrainst, this will
           return a list of pairs such that for each of base and snap
           declarations, well expand like so (showing on the base declaration
           for simplicity):

               base = {
                   'slots': {
                       'interface': {
                           'foo': '1',
                           'bar': ['2', '3'],
                           'baz': '4',
                           'norf': ['5', '6'],
                       }
                   },
                   'plugs': {
                       'interface': {
                           'qux': '7',
                           'quux': ['8', '9'],
                       }
                   }
               }

            then the list of 'base declarations' to check against is:

                decls['base'] = [
                    {'slots': {
                        'interface': {
                            'foo': '1',
                            'bar': '2',
                            'baz': '4',
                            'norf': '5',
                        },
                    },
                    {'slots': {
                        'interface': {
                            'foo': '1',
                            'bar': '2',
                            'baz': '4',
                            'norf': '6',
                        }
                    },
                    {'slots': {
                        'interface': {
                            'foo': '1',
                            'bar': '3',
                            'baz': '4',
                            'norf': '5',
                        }
                    },
                    {'slots': {
                        'interface': {
                            'foo': '1',
                            'bar': '3',
                            'baz': '4',
                            'norf': '6',
                        }
                    },
                    {'plugs': {
                        'interface': {
                            'qux': '7',
                            'quux': '8',
                        }
                    },
                    {'plugs': {
                        'interface': {
                            'qux': '7',
                            'quux': '9',
                        }
                    },
                ]

            If the plugs side is defined for this interface, it will appear
            next to the slot as with a regular declaration. If the snap
            declaration is defined, it will be stored in decls['snap'] in the
            same way as the base declaration.

            In this manner, each one of the base declarations can be evaluated
            and compared to any defined snap declarations.
        '''
        def expand(d, side, interface, keys, templates):
            if len(keys) == 0:
                return templates

            updated = []
            key = keys[-1]
            for i in d[side][interface][key]:
                for t in templates:
                    tmp = {side: {interface: {}}}
                    # copy existing keys
                    for template_key in t[side][interface]:
                        tmp[side][interface][template_key] = \
                            t[side][interface][template_key]
                    tmp[side][interface][key] = i
                    updated.append(tmp)

            return expand(d, side, interface, keys[:-1], updated)

        decls = {'base': [], 'snap': []}

        has_alternates = False
        for dtype in ["base", "snap"]:
            if dtype == "base":
                d = self.base_declaration
            else:
                d = self.snap_declaration

            tmp = {}
            for side in ["plugs", "slots"]:
                if dtype == "snap" and d is None:
                    continue
                if side not in d or interface not in d[side]:
                    continue

                to_expand = []
                template = {side: {interface: {}}}
                for cstr in d[side][interface]:
                    if isinstance(d[side][interface][cstr], list):
                        to_expand.append(cstr)
                    else:
                        template[side][interface][cstr] = \
                            d[side][interface][cstr]

                tmp[side] = []
                tmp[side] += expand(d, side, interface, to_expand, [template])

                if len(to_expand) > 0:
                    has_alternates = True

            # Now that we have all the slots combinations and all the plugs
            # combinations, create combinations of those
            if "plugs" in tmp and "slots" in tmp:
                for p in tmp["plugs"]:
                    for s in tmp["slots"]:
                        decls[dtype].append({'plugs': p['plugs'],
                                             'slots': s['slots']})
            elif "plugs" in tmp:
                decls[dtype] = tmp["plugs"]
            elif "slots" in tmp:
                decls[dtype] = tmp["slots"]

        # We need at least one declaration per list, even if it is None
        if len(decls['snap']) == 0:
            decls['snap'].append(None)

        return (decls, has_alternates)

    def _verify_iface_by_declaration(self, base, snap, name, iface, interface,
                                     attribs, side, oside):
        # 'checked' is used to see if a particular check is made (eg, if
        # 'deny-connection' for this interface was performed).
        #
        # 'denied' is used to track if something checked prompted manual review
        #
        # _verify_iface_by_declaration() will return if something prompted
        # manual review (denied > 0) and if this is an exact match (ie, if
        # checked == denied).
        checked = 0
        denied = 0

        def err(key, subkey=None, dtype="base", attrs=None):
            s = "human review required due to '%s' constraint " % key
            if subkey is not None:
                s += "for '%s' " % subkey
            s += "from %s declaration" % dtype

            if attrs is not None:
                if 'allow-sandbox' in attrs and attrs['allow-sandbox']:
                    s += ". If using a chromium webview, you can disable " + \
                         "the internal sandbox (eg, use --no-sandbox) and " + \
                         "remove the 'allow-sandbox' attribute instead. " + \
                         "For Oxide webviews, export OXIDE_NO_SANDBOX=1 " + \
                         "to disable its internal sandbox."

            return s

        # top-level allow/deny-installation/connection
        # Note: auto-connection is only for snapd, so don't include it here
        for i in ['installation', 'connection']:
            for j in ['deny', 'allow']:
                decl_key = "%s-%s" % (j, i)
                # flag if deny-* is true or allow-* is false
                (decl, base_decl, decl_type) = self._get_decl(base, snap, side,
                                                              interface, i)
                if side in decl and interface in decl[side] and \
                        decl_key in decl[side][interface] and \
                        not isinstance(decl[side][interface][decl_key], dict):
                    checked += 1
                    if self._search(decl[side][interface], decl_key,
                                    j == 'deny'):
                        self._add_result('error',
                                         self._get_check_name("%s_%s" %
                                                              (side, decl_key),
                                                              app=iface,
                                                              extra=interface),
                                         err(decl_key, dtype=decl_type),
                                         manual_review=True,
                                         stage=True)
                        denied += 1

                        # if manual review after 'deny', don't look at allow
                        break

        # deny/allow-installation snap-type
        snap_type = 'app'
        if 'type' in self.snap_yaml:
            snap_type = self.snap_yaml['type']
            if snap_type == 'os':
                snap_type = 'core'
        decl_subkey = '%s-snap-type' % side[:-1]
        for j in ['deny', 'allow']:
            (decl, base_decl, decl_type) = self._get_decl(base, snap, side,
                                                          interface,
                                                          'installation')
            decl_key = "%s-installation" % j
            # flag if deny-*/snap-type matches or allow-*/snap-type doesn't
            if side in decl and interface in decl[side] and \
                    decl_key in decl[side][interface] and \
                    isinstance(decl[side][interface][decl_key], dict) and \
                    decl_subkey in decl[side][interface][decl_key]:
                checked += 1
                if self._search(decl[side][interface], decl_key,
                                subkey=decl_subkey, subval=snap_type,
                                subval_inverted=(j == 'allow')):
                    self._add_result('error',
                                     self._get_check_name("%s_%s" %
                                                          (side, decl_key),
                                                          app=iface,
                                                          extra=interface),
                                     err(decl_key, decl_subkey, decl_type),
                                     manual_review=True,
                                     stage=True)
                    denied += 1

                    # if manual review after 'deny', don't look at allow
                    break

        # deny/allow-connection/installation on-classic with app snaps
        # Note: auto-connection is only for snapd, so don't include it here
        snap_type = 'app'
        if 'type' in self.snap_yaml:
            snap_type = self.snap_yaml['type']
            if snap_type == 'os':
                snap_type = 'core'
        decl_subkey = 'on-classic'
        for i in ['installation', 'connection']:
            for j in ['deny', 'allow']:
                (decl, base_decl, decl_type) = self._get_decl(base, snap, side,
                                                              interface, i)
                decl_key = "%s-%s" % (j, i)
                # when an app snap, flag if deny-*/on-classic=false or
                # allow-*/on-classic=true
                # when not an app snap, flag if deny-*/on-classic=true or
                # allow-*/on-classic=false
                if side in decl and interface in decl[side] and \
                        decl_key in decl[side][interface] and \
                        isinstance(decl[side][interface][decl_key], dict) and \
                        decl_subkey in decl[side][interface][decl_key]:
                    checked += 1
                    if self._search(decl[side][interface], decl_key,
                                    subkey=decl_subkey,
                                    subval=(snap_type == 'app'),
                                    subval_inverted=(j == 'deny')):
                        self._add_result('error',
                                         self._get_check_name("%s_%s" %
                                                              (side, decl_key),
                                                              app=iface,
                                                              extra=interface),
                                         err(decl_key, decl_subkey, decl_type),
                                         manual_review=True,
                                         stage=True)
                        denied += 1

                        # if manual review after 'deny', don't look at allow
                        break

        # deny/allow-connection/installation attributes
        # Note: auto-connection is only for snapd, so don't include it here
        decl_subkey = '%s-attributes' % side[:-1]
        for i in ['installation', 'connection']:
            if attribs is None:
                continue
            for j in ['deny', 'allow']:
                (decl, base_decl, decl_type) = self._get_decl(base, snap, side,
                                                              interface, i)
                decl_key = "%s-%s" % (j, i)
                # flag if any deny-*/attribs match or any allow-*/attribs don't
                if side in decl and interface in decl[side] and \
                        decl_key in decl[side][interface] and \
                        isinstance(decl[side][interface][decl_key], dict) and \
                        decl_subkey in decl[side][interface][decl_key]:
                    checked += 1
                    if self._search(decl[side][interface], decl_key,
                                    subkey=decl_subkey, subval=attribs,
                                    subval_inverted=(j == 'allow')):
                        self._add_result('error',
                                         self._get_check_name("%s_%s" %
                                                              (side, decl_key),
                                                              app=iface,
                                                              extra=interface),
                                         err(decl_key, decl_subkey,
                                             decl_type, attribs),
                                         manual_review=True,
                                         stage=True)
                        denied += 1

                        # if manual review after 'deny', don't look at allow
                        break
                # Since base declaration mostly has slots side, if plugs, look
                # at the other side for checking plug-attributes
                elif base_decl and side == 'plugs' and oside in decl and \
                        interface in decl[oside] and \
                        decl_key in decl[oside][interface] and \
                        decl_subkey in decl[oside][interface][decl_key]:
                    checked += 1
                    if self._search(decl[oside][interface], decl_key,
                                    subkey=decl_subkey, subval=attribs,
                                    subval_inverted=(j == 'allow')):
                        self._add_result('error',
                                         self._get_check_name("%s_%s" %
                                                              (side, decl_key),
                                                              app=iface,
                                                              extra=interface),
                                         err(decl_key, decl_subkey,
                                             decl_type, attribs),
                                         manual_review=True,
                                         stage=True)
                        denied += 1

                        # if manual review after 'deny', don't look at allow
                        break

        # Return if something prompted for manual review and if everything
        # checked was denied (an exact match denial)
        return (denied > 0, checked == denied)

    def _verify_iface(self, name, iface, interface, attribs=None):
        if name.endswith('slot'):
            side = 'slots'
            oside = 'plugs'
        elif name.endswith('plug'):
            side = 'plugs'
            oside = 'slots'

        t = 'info'
        n = self._get_check_name('%s_known' % name, app=iface, extra=interface)
        s = 'OK'
        if side in self.base_declaration and \
                interface not in self.base_declaration[side] and \
                oside in self.base_declaration and \
                interface not in self.base_declaration[oside]:
            if name.startswith('app_') and side in self.snap_yaml and \
                    interface in self.snap_yaml[side]:
                # If it is an interface reference used by an app, skip since it
                # will be checked in top-level interface checks.
                return
            t = 'error'
            s = "interface '%s' not found in base declaration" % interface
            self._add_result(t, n, s)
            return

        # To support alternates in the base and snap declaration, we need to
        # try each combination of snap alternate constraint and base alternate
        # constraint. If we have alternates and one passes and there are no
        # exact denials, then don't report. Otherwise report if require manual
        # review.
        (decls, has_alternates) = self._get_all_combinations(interface)
        require_manual = False

        exact_deny = True
        for b in decls['base']:
            for s in decls['snap']:
                (manual, exact) = \
                    self._verify_iface_by_declaration(b, s, name, iface,
                                                      interface, attribs, side,
                                                      oside)
                if manual:
                    require_manual = True
                    if has_alternates and not exact:
                        exact_deny = False

        if has_alternates and not exact_deny:
            require_manual = False

        # Apply our staged results if required, otherwise report all is ok
        if require_manual:
            self._apply_staged_results()
        else:
            self._add_result('info',
                             self._get_check_name("%s" % side, app=iface,
                                                  extra=interface),
                             "OK", manual_review=False)

    def check_declaration(self):
        '''Check base/snap declaration requires manual review for top-level
           plugs/slots
        '''
        if not self.is_snap2:
            return

        for side in ['plugs', 'slots']:
            if side not in self.snap_yaml:
                continue

            for iface in self.snap_yaml[side]:
                # If the 'interface' name is the same as the 'plug/slot' name,
                # then 'interface' is optional since the interface name and the
                # plug/slot name are the same
                interface = iface
                attribs = None

                spec = self.snap_yaml[side][iface]
                if isinstance(spec, str):
                    # Abbreviated syntax (no attributes)
                    # <plugs|slots>:
                    #   <alias>: <interface>
                    interface = spec
                elif 'interface' in spec:
                    # Full specification.
                    # <plugs|slots>:
                    #   <alias>:
                    #     interface: <interface>
                    interface = spec['interface']
                    if len(spec) > 1:
                        attribs = spec
                        del attribs['interface']

                self._verify_iface(side[:-1], iface, interface, attribs)

    def check_declaration_apps(self):
        '''Check base/snap declaration requires manual review for apps
           plugs/slots
        '''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            for side in ['plugs', 'slots']:
                if side not in self.snap_yaml['apps'][app]:
                    continue

                # The interface referenced in the app's 'plugs' or 'slots'
                # field can either be a known interface (when the interface
                # name reference and the interface is the same) or can
                # reference a name in the snap's toplevel 'plugs' or 'slots'
                # mapping
                for ref in self.snap_yaml['apps'][app][side]:
                    if not isinstance(ref, str):
                        continue  # checked elsewhere

                    self._verify_iface('app_%s' % side[:-1], app, ref)



'''sr_lint.py: lint checks'''
#
# Copyright (C) 2013-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
from clickreviews.sr_common import (
    SnapReview,
)
from clickreviews.common import (
    find_external_symlinks,
    STORE_PKGNAME_SNAPV2_MAXLEN,
)
from clickreviews.overrides import (
    redflagged_snap_types_overrides,
    desktop_file_exception
)
import glob
import os
import re


class SnapReviewLint(SnapReview):
    '''This class represents snap lint reviews'''

    def __init__(self, fn, overrides=None):
        '''Set up the class.'''
        SnapReview.__init__(self, fn, "lint-snap-v2", overrides=overrides)
        if not self.is_snap2:
            return

        self.valid_compiled_architectures = ['armhf',
                                             'i386',
                                             'amd64',
                                             'arm64',
                                             'powerpc',
                                             'ppc64el',
                                             's390x',
                                             ]
        self.valid_architectures = ['all'] + self.valid_compiled_architectures
        self.vcs_files = ['.bzr*',
                          # '.excludes',  # autogenerated by SDK
                          '.git*',
                          '.idea',
                          '.svn*',
                          '.hg',
                          '.project',
                          'CVS*',
                          'RCS*'
                          ]
        self.iffy_files = [
            r'^\..+\.swp$',  # vim
        ]

        self._list_all_compiled_binaries()

        self.redflagged_snap_types = ['kernel',
                                      'gadget',
                                      'os',
                                      ]

        self.interface_plug_requires_desktop_file = ['unity7',
                                                     'x11',
                                                     'unity8'
                                                     ]

    def check_architectures(self):
        '''Check architectures in snap.yaml is valid'''
        if not self.is_snap2:
            return

        t = 'info'
        n = self._get_check_name('architecture_valid')
        s = 'OK'

        key = 'architectures'
        if key not in self.snap_yaml:
            s = 'OK (%s not specified)' % key
            self._add_result(t, n, s)
            return

        if not isinstance(self.snap_yaml[key], list):
            t = 'error'
            s = "invalid %s entry: %s (not a list)" % (key,
                                                       self.snap_yaml[key])
        else:
            bad_archs = []
            for arch in self.snap_yaml[key]:
                if not isinstance(arch, str):
                    bad_archs.append(str(arch))
                elif arch not in self.valid_architectures:
                    bad_archs.append(arch)

                if len(bad_archs) > 0:
                    t = 'error'
                    s = "invalid multi architecture: %s" % ",".join(bad_archs)
        self._add_result(t, n, s)

    def check_assumes(self):
        '''Check assumes in snap.yaml is valid'''
        if not self.is_snap2:
            return

        t = 'info'
        n = self._get_check_name('assumes_valid')
        s = 'OK'

        key = 'assumes'
        if key not in self.snap_yaml:
            s = 'OK (%s not specified)' % key
            self._add_result(t, n, s)
            return

        if not isinstance(self.snap_yaml[key], list):
            t = 'error'
            s = "invalid %s entry: %s (not a list)" % (key,
                                                       self.snap_yaml[key])
        else:
            bad_assumes = []
            for a in self.snap_yaml[key]:
                if not isinstance(a, str):
                    bad_assumes.append(str(a))
            if len(bad_assumes) > 0:
                t = 'error'
                s = "invalid assumes: %s" % ",".join(bad_assumes)
        self._add_result(t, n, s)

    def check_description(self):
        '''Check description'''
        if not self.is_snap2:
            return

        key = 'description'

        t = 'info'
        n = self._get_check_name('%s_present' % key)
        s = 'OK'
        if key not in self.snap_yaml:
            s = 'OK (optional %s field not specified)' % key
            self._add_result(t, n, s)
            return
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name(key)
        s = 'OK'
        if not isinstance(self.snap_yaml[key], str):
            t = 'error'
            s = "invalid %s entry: %s (not a str)" % (key, self.snap_yaml[key])
            self._add_result(t, n, s)
            return
        elif len(self.snap_yaml[key]) < 1:
            t = 'error'
            s = "invalid %s entry (empty)" % (key)
        elif len(self.snap_yaml[key]) < len(self.snap_yaml['name']):
            t = 'info'
            s = "%s is too short: '%s'" % (key, self.snap_yaml[key])
        self._add_result(t, n, s)

    def check_name(self):
        '''Check package name'''
        if not self.is_snap2:
            return

        t = 'info'
        n = self._get_check_name('name_valid')
        s = 'OK'
        if 'name' not in self.snap_yaml:
            t = 'error'
            s = "could not find 'name' in yaml"
        elif not isinstance(self.snap_yaml['name'], str):
            t = 'error'
            s = "malformed 'name': %s (not a str)" % (self.snap_yaml['name'])
        elif len(self.snap_yaml['name']) > STORE_PKGNAME_SNAPV2_MAXLEN:
            t = 'error'
            s = "malformed 'name': '%s' " % self.snap_yaml['name'] + \
                "(length '%d' " % len(self.snap_yaml['name']) + \
                "exceeds store limit '%d')" % STORE_PKGNAME_SNAPV2_MAXLEN
        elif not self._verify_pkgname(self.snap_yaml['name']):
            t = 'error'
            s = "malformed 'name': '%s' " % self.snap_yaml['name'] + \
                "(may be only lower case, digits and hyphens)"
        self._add_result(t, n, s)

    def check_summary(self):
        '''Check summary'''
        if not self.is_snap2:
            return

        key = 'summary'

        t = 'info'
        n = self._get_check_name('%s_present' % key)
        s = 'OK'
        if key not in self.snap_yaml:
            s = 'OK (optional %s field not specified)' % key
            self._add_result(t, n, s)
            return
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name(key)
        s = 'OK'
        if not isinstance(self.snap_yaml[key], str):
            t = 'error'
            s = "invalid %s entry: %s (not a str)" % (key, self.snap_yaml[key])
            self._add_result(t, n, s)
            return
        elif len(self.snap_yaml[key]) < 1:
            t = 'error'
            s = "invalid %s entry (empty)" % (key)
        elif len(self.snap_yaml[key]) < len(self.snap_yaml['name']):
            t = 'info'
            s = "%s is too short: '%s'" % (key, self.snap_yaml[key])
        self._add_result(t, n, s)

    def check_type(self):
        '''Check type'''
        if not self.is_snap2 or 'type' not in self.snap_yaml:
            return

        t = 'info'
        n = self._get_check_name('snap_type_valid')
        s = 'OK'
        if self.snap_yaml['type'] not in self.valid_snap_types:
            t = 'error'
            s = "unknown 'type': '%s'" % self.snap_yaml['type']
        self._add_result(t, n, s)

    def check_type_redflagged(self):
        '''Check if type is redflagged'''
        if not self.is_snap2 or 'type' not in self.snap_yaml:
            return

        t = 'info'
        n = self._get_check_name('snap_type_redflag')
        s = "OK"
        manual_review = False
        if self.snap_yaml['type'] in self.redflagged_snap_types:
            pkgname = self.snap_yaml['name']
            snaptype = self.snap_yaml['type']
            if snaptype in redflagged_snap_types_overrides and \
                    pkgname in redflagged_snap_types_overrides[snaptype]:
                s = "OK (override '%s' for 'type: %s')" % (pkgname,
                                                           snaptype)
            else:
                t = 'error'
                s = "(NEEDS REVIEW) type '%s' not allowed" % \
                    self.snap_yaml['type']
                manual_review = True
        self._add_result(t, n, s, manual_review=manual_review)

    def check_version(self):
        '''Check package version'''
        if not self.is_snap2:
            return

        t = 'info'
        n = self._get_check_name('version_valid')
        s = 'OK'
        if 'version' not in self.snap_yaml:
            t = 'error'
            s = "could not find 'version' in yaml"
        elif not self._verify_pkgversion(self.snap_yaml['version']):
            t = 'error'
            s = "malformed 'version': '%s'" % self.snap_yaml['version']
        self._add_result(t, n, s)

    def check_config(self):
        '''Check config'''
        if not self.is_snap2:
            return

        fn = os.path.join(self._get_unpack_dir(), 'meta/hooks/config')
        if fn not in self.pkg_files:
            return

        t = 'info'
        n = self._get_check_name('config_hook_executable')
        s = 'OK'
        if not self._check_innerpath_executable(fn):
            t = 'error'
            s = 'meta/hooks/config is not executable'
        self._add_result(t, n, s)

    def check_icon(self):
        '''Check icon'''
        # see docs/meta.md and docs/gadget.md
        if not self.is_snap2 or 'icon' not in self.snap_yaml:
            return

        # Snappy icons may be specified in the gadget snap.yaml, but not in
        # app snap.yaml. With apps, the icon may be in meta/gui/icon.png and
        # this file is optional. Therefore, for apps, there is nothing to do.
        t = 'info'
        n = self._get_check_name('icon_present')
        s = 'OK'
        if 'type' in self.snap_yaml and self.snap_yaml['type'] != "gadget":
            t = 'warn'
            s = 'icon only used with gadget snaps'
            self._add_result(t, n, s)
            return
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('icon_empty')
        s = 'OK'
        if len(self.snap_yaml['icon']) == 0:
            t = 'error'
            s = "icon entry is empty"
            self._add_result(t, n, s)
            return
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('icon_absolute_path')
        s = 'OK'
        if self.snap_yaml['icon'].startswith('/'):
            t = 'error'
            s = "icon entry '%s' should not specify absolute path" % \
                self.snap_yaml['icon']
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('icon_exists')
        s = 'OK'
        fn = self._path_join(self._get_unpack_dir(), self.snap_yaml['icon'])
        if fn not in self.pkg_files:
            t = 'error'
            s = "icon entry '%s' does not exist" % self.snap_yaml['icon']
        self._add_result(t, n, s)

    def check_unknown_entries(self):
        '''Check for any unknown fields'''
        if not self.is_snap2:
            return

        t = 'info'
        n = self._get_check_name('unknown_field')
        s = 'OK'
        unknown = []
        for f in self.snap_yaml:
            if f not in self.snappy_required + self.snappy_optional:
                unknown.append(f)
        if len(unknown) > 0:
            t = 'warn'
            s = "unknown entries in snap.yaml: '%s'" % \
                (",".join(sorted(unknown)))
        self._add_result(t, n, s)

    def _verify_apps_and_hooks(self, hook=False):
        key = 'apps'
        key_type = 'app'
        required = self.apps_required
        optional = self.apps_optional
        if hook:
            key = 'hooks'
            key_type = 'hook'
            required = self.hooks_required
            optional = self.hooks_optional

        t = 'info'
        n = self._get_check_name('%s_present' % key)
        s = 'OK'
        if key not in self.snap_yaml:
            s = 'OK (optional %s field not specified)' % key
            self._add_result(t, n, s)
            return
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name(key)
        s = 'OK'
        if not isinstance(self.snap_yaml[key], dict):
            t = 'error'
            s = "invalid %s entry: %s (not a dict)" % (key,
                                                       self.snap_yaml[key])
            self._add_result(t, n, s)
            return
        elif len(self.snap_yaml[key].keys()) < 1:
            t = 'error'
            s = "invalid %s entry (empty)" % (key)
            self._add_result(t, n, s)
            return
        self._add_result(t, n, s)

        for val in self.snap_yaml[key]:
            t = 'info'
            n = self._get_check_name('%s_entry' % key, app=val)
            s = 'OK'

            if not isinstance(self.snap_yaml[key][val], dict):
                if key_type == 'hook' and self.snap_yaml[key][val] is None:
                    s = "OK (hook entry is empty)"
                else:
                    t = 'error'
                    s = "invalid entry: %s (not a dict)" % (
                        self.snap_yaml[key][val])
                self._add_result(t, n, s)
                continue
            elif key_type == 'app' and \
                    len(self.snap_yaml[key][val].keys()) < 1:
                t = 'error'
                s = "invalid entry for '%s' (empty)" % (val)
                self._add_result(t, n, s)
                continue
            elif not self._verify_appname(val):
                t = 'error'
                s = "malformed %s name: '%s'" % (key_type, val)
                self._add_result(t, n, s)
                continue
            self._add_result(t, n, s)

            for field in required:
                t = 'info'
                n = self._get_check_name('%s_required' % key, app=val)
                s = 'OK'
                if field not in self.snap_yaml[key][val]:
                    t = 'error'
                    s = "required field '%s' not specified" % field
                self._add_result(t, n, s)

            t = 'info'
            n = self._get_check_name('%s_unknown' % key, app=val)
            s = 'OK'
            unknown = []
            for field in self.snap_yaml[key][val]:
                if field not in required + optional:
                    unknown.append(field)
            if len(unknown) > 0:
                t = 'warn'
                s = "unknown fields for %s '%s': '%s'" % (
                    key_type, val, ",".join(sorted(unknown)))
            self._add_result(t, n, s)

    def check_apps(self):
        '''Check apps'''
        if not self.is_snap2:
            return
        self._verify_apps_and_hooks()

    def check_hooks(self):
        '''Check hooks'''
        if not self.is_snap2:
            return
        self._verify_apps_and_hooks(hook=True)

    def _verify_value_is_file(self, app, key):
        t = 'info'
        n = self._get_check_name('%s' % key, app=app)
        s = 'OK'
        if not isinstance(self.snap_yaml['apps'][app][key], str):
            t = 'error'
            s = "%s '%s' (not a str)" % (key,
                                         self.snap_yaml['apps'][app][key])
            self._add_result(t, n, s)
        elif len(self.snap_yaml['apps'][app][key]) < 1:
            t = 'error'
            s = "invalid %s (empty)" % (key)
            self._add_result(t, n, s)
        else:
            fn = self._path_join(self._get_unpack_dir(),
                                 os.path.normpath(
                                     self.snap_yaml['apps'][app][key]))
            if fn not in self.pkg_files:
                t = 'error'
                s = "%s does not exist" % (
                    self.snap_yaml['apps'][app][key])
        self._add_result(t, n, s)

    def check_apps_command(self):
        '''Check apps - command'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'command'
            if key not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            self._verify_value_is_file(app, key)

    def check_apps_stop_command(self):
        '''Check apps - stop-command'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'stop-command'
            if key not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            self._verify_value_is_file(app, key)

    def check_apps_post_stop_command(self):
        '''Check apps - post-stop-command'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'post-stop-command'
            if key not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            self._verify_value_is_file(app, key)

    def check_apps_stop_timeout(self):
        '''Check apps - stop-timeout'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'stop-timeout'
            if key not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            suffix = 'ns|us|ms|s|m'
            t = 'info'
            n = self._get_check_name('%s' % key, app=app)
            s = "OK"
            if not isinstance(self.snap_yaml['apps'][app][key], int) and \
                    not isinstance(self.snap_yaml['apps'][app][key], str):
                t = 'error'
                s = "'%s' is not a string or integer" % key
            elif not re.search(r'[0-9]+(%s)?$' % suffix,
                               str(self.snap_yaml['apps'][app][key])):
                t = 'error'
                s = "'%s' is not of form NN[ms] (%s)" % \
                    (self.snap_yaml['apps'][app][key], key)
            self._add_result(t, n, s)

            if t == 'error':
                continue

            t = 'info'
            n = self._get_check_name('%s_range' % key, app=app)
            s = "OK"
            st = int(str(self.snap_yaml['apps'][app][key]).rstrip(r'(%s)' %
                                                                  suffix))
            if st < 0:
                t = 'error'
                s = "stop-timeout '%s' should be a positive" % \
                    str(self.snap_yaml['apps'][app][key])
            self._add_result(t, n, s)

    def _verify_valid_values(self, app, key, valid):
        '''Verify valid values for key in app'''
        t = 'info'
        n = self._get_check_name('%s' % key, app=app)
        s = 'OK'
        if not isinstance(self.snap_yaml['apps'][app][key], str):
            t = 'error'
            s = "%s '%s' (not a str)" % (key,
                                         self.snap_yaml['apps'][app][key])
            self._add_result(t, n, s)
        elif len(self.snap_yaml['apps'][app][key]) < 1:
            t = 'error'
            s = "invalid %s (empty)" % (key)
            self._add_result(t, n, s)
        elif self.snap_yaml['apps'][app][key] not in valid:
            t = 'error'
            s = "invalid %s: '%s'" % (key, self.snap_yaml['apps'][app][key])
        self._add_result(t, n, s)

    def check_apps_daemon(self):
        '''Check apps - daemon'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        valid = ["simple",
                 "forking",
                 "oneshot",
                 "notify",
                 ]

        for app in self.snap_yaml['apps']:
            key = 'daemon'
            if key not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            self._verify_valid_values(app, key, valid)

    def check_apps_nondaemon(self):
        '''Check apps - non-daemon'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        # Certain options require 'daemon' so list the keys that are shared
        # by services and binaries
        ok_keys = ['command', 'completer', 'environment', 'plugs', 'slots',
                   'aliases']

        for app in self.snap_yaml['apps']:
            needs_daemon = []
            for key in self.snap_yaml['apps'][app]:
                if key not in self.apps_optional or \
                        key == 'daemon' or \
                        key in ok_keys or \
                        'daemon' in self.snap_yaml['apps'][app]:
                    continue
                needs_daemon.append(key)

            t = 'info'
            n = self._get_check_name('daemon_required', app=app)
            s = "OK"
            if len(needs_daemon) > 0:
                t = 'error'
                s = "'%s' must be used with 'daemon'" % ",".join(needs_daemon)
            self._add_result(t, n, s)

    def check_apps_restart_condition(self):
        '''Check apps - restart-condition'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        valid = ["always",
                 "never",
                 "on-abnormal",
                 "on-abort",
                 "on-failure",
                 "on-success",
                 ]

        for app in self.snap_yaml['apps']:
            key = 'restart-condition'
            if key not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            self._verify_valid_values(app, key, valid)

    def check_apps_ports(self):
        '''Check apps - ports'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        valid_keys = ['internal', 'external']
        valid_subkeys = ['port', 'negotiable']
        for app in self.snap_yaml['apps']:
            if 'ports' not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            t = 'info'
            n = self._get_check_name('ports', app=app)
            s = 'OK'
            link = None
            if not isinstance(self.snap_yaml['apps'][app]['ports'], dict):
                t = 'error'
                s = "ports '%s' (not a dict)" % (
                    self.snap_yaml['apps'][app]['ports'])
            elif len(self.snap_yaml['apps'][app]['ports'].keys()) < 1:
                t = 'error'
                s = "'ports' must contain 'internal' and/or 'external'"
            self._add_result(t, n, s, link)
            if t == 'error':
                continue

            # unknown
            unknown = []
            for key in self.snap_yaml['apps'][app]['ports']:
                if key not in valid_keys:
                    unknown.append(key)
            if len(unknown) > 0:
                t = 'error'
                n = self._get_check_name('ports_unknown_key', extra=key,
                                         app=app)
                s = "Unknown '%s' for ports" % (",".join(unknown))
                self._add_result(t, n, s)

            port_pat = re.compile(r'^[0-9]+/[a-z0-9\-]+$')
            for key in valid_keys:
                if key not in self.snap_yaml['apps'][app]['ports']:
                    continue

                if len(self.snap_yaml['apps'][app]['ports'][key].keys()) < 1:
                    t = 'error'
                    n = self._get_check_name('ports', extra=key, app=app)
                    s = 'Could not find any %s ports' % key
                    self._add_result(t, n, s)
                    continue

                for tagname in self.snap_yaml['apps'][app]['ports'][key]:
                    entry = self.snap_yaml['apps'][app]['ports'][key][tagname]
                    if len(entry.keys()) < 1 or ('negotiable' not in entry and
                                                 'port' not in entry):
                        t = 'error'
                        n = self._get_check_name('ports', extra=key, app=app)
                        s = "Could not find 'port' or 'negotiable' in '%s'" % \
                            tagname
                        self._add_result(t, n, s)
                        continue

                    # unknown
                    unknown = []
                    for subkey in entry:
                        if subkey not in valid_subkeys:
                            unknown.append(subkey)
                    if len(unknown) > 0:
                        t = 'error'
                        n = self._get_check_name('ports_unknown_subkey',
                                                 extra=key, app=app)
                        s = "Unknown '%s' for %s" % (",".join(unknown),
                                                     tagname)
                        self._add_result(t, n, s)

                    # port
                    subkey = 'port'
                    t = 'info'
                    n = self._get_check_name('ports_%s_format' % tagname,
                                             extra=subkey)
                    s = 'OK'
                    if subkey not in entry:
                        s = 'OK (skipped, not found)'
                    elif not isinstance(entry[subkey], str):
                        t = 'error'
                        s = "invalid entry: %s (not a str)" % (entry[subkey])
                    else:
                        tmp = entry[subkey].split('/')
                        if not port_pat.search(entry[subkey]) or \
                           int(tmp[0]) < 1 or int(tmp[0]) > 65535:
                            t = 'error'
                            s = "'%s' should be of form " % entry[subkey] + \
                                "'port/protocol' where port is an integer " + \
                                "(1-65535) and protocol is found in " + \
                                "/etc/protocols"
                    self._add_result(t, n, s)

                    # negotiable
                    subkey = 'negotiable'
                    t = 'info'
                    n = self._get_check_name('ports_%s_format' % tagname,
                                             extra=subkey)
                    s = 'OK'
                    if subkey not in entry:
                        s = 'OK (skipped, not found)'
                    elif not isinstance(entry[subkey], bool):
                        t = 'error'
                        s = "'%s: %s' should be either 'yes' or 'no'" % \
                            (subkey, entry[subkey])
                    self._add_result(t, n, s)

    def check_apps_socket(self):
        '''Check apps - socket'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'socket'
            if key not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            t = 'info'
            n = self._get_check_name(key, app=app)
            s = 'OK'
            if not isinstance(self.snap_yaml['apps'][app][key], bool):
                t = 'error'
                s = "'%s: %s' should be either 'yes' or 'no'" % (
                    key, self.snap_yaml['apps'][app][key])
            elif 'listen-stream' not in self.snap_yaml['apps'][app]:
                t = 'error'
                s = "'socket' specified without 'listen-stream'"
            self._add_result(t, n, s)

    def check_apps_listen_stream(self):
        '''Check apps - listen-stream'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'listen-stream'
            if key not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            t = 'info'
            n = self._get_check_name(key, app=app)
            s = 'OK'
            if not isinstance(self.snap_yaml['apps'][app][key], str):
                t = 'error'
                s = "invalid entry: %s (not a str)" % (
                    self.snap_yaml['apps'][app][key])
            elif len(self.snap_yaml['apps'][app][key]) == 0:
                t = 'error'
                s = "'%s' is empty" % key
            self._add_result(t, n, s)
            if t == 'error':
                continue

            t = 'info'
            n = self._get_check_name('%s_matches_name' % key, app=app)
            s = 'OK'
            sock = self.snap_yaml['apps'][app][key]
            pkgname = self.snap_yaml['name']
            if sock.startswith('@'):
                if sock != '@%s' % pkgname and \
                        not sock.startswith('@%s_' % pkgname):
                    t = 'error'
                    s = ("abstract socket '%s' is neither '%s' nor starts "
                         "with '%s'" % (sock, '@%s' % pkgname,
                                        '@%s_' % pkgname))
            elif sock.startswith('/'):
                found = False
                for path in ["/tmp/",
                             "/var/lib/snaps/%s/" % pkgname,
                             "/var/lib/snaps/%s." % pkgname,
                             "/run/shm/snaps/%s/" % pkgname,
                             "/run/shm/snaps/%s." % pkgname]:
                    if sock.startswith(path):
                        found = True
                        break
                if not found:
                    t = 'error'
                    s = ("named socket '%s' should be in a writable "
                         "app-specific area or /tmp" % sock)
            else:
                t = 'error'
                s = ("'%s' does not specify an abstract socket (starts "
                     "with '@') or absolute filename" % (sock))
            self._add_result(t, n, s)

    def _verify_valid_socket(self, app, key):
        '''Verify valid values for socket key'''
        t = 'info'
        n = self._get_check_name(key, app=app)
        s = 'OK'
        if not isinstance(self.snap_yaml['apps'][app][key], str):
            t = 'error'
            s = "invalid entry: %s (not a str)" % (
                self.snap_yaml['apps'][app][key])
        elif len(self.snap_yaml['apps'][app][key]) == 0:
            t = 'error'
            s = "'%s' is empty" % key
        elif 'listen-stream' not in self.snap_yaml['apps'][app]:
            t = 'error'
            s = "'%s' specified without 'listen-stream'" % key
        self._add_result(t, n, s)
        if t == 'error':
            return

        t = 'error'
        n = self._get_check_name('%s_reserved' % key, app=app)
        s = "'%s' should not be used until snappy supports per-app users" \
            % key
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name("%s_matches_name" % key, app=app)
        s = 'OK'
        if self.snap_yaml['apps'][app][key] != self.snap_yaml['name']:
            t = 'error'
            s = "'%s' != '%s'" % (self.snap_yaml['apps'][app][key],
                                  self.snap_yaml['name'])
        self._add_result(t, n, s)

    def check_apps_socket_user(self):
        '''Check apps - socket-user'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'socket-user'
            if key not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            self._verify_valid_socket(app, key)

    def check_apps_socket_group(self):
        '''Check apps - socket-group'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'socket-group'
            if key not in self.snap_yaml['apps'][app]:
                # We check for required elsewhere
                continue

            self._verify_valid_socket(app, key)

    def _verify_interfaces(self, iface_type):
        for iface in self.snap_yaml[iface_type]:
            # If the 'interface' name is the same as the 'iface' name, then
            # 'interface' is optional since the interface name and the iface
            # name are the same
            interface = iface

            spec = self.snap_yaml[iface_type][iface]
            if isinstance(spec, str):
                # Abbreviated syntax (no attributes)
                # <plugs|slots>:
                #   <alias>: <interface>
                interface = spec
            elif 'interface' in spec:
                # Full specification.
                # <plugs|slots>:
                #   <alias>:
                #     interface: <interface>
                interface = spec['interface']

            # Validate indirect (specified via alias) interfaces:
            if interface != iface:
                key = 'interface'
                t = 'info'
                n = self._get_check_name(iface_type, app=key, extra=interface)
                s = 'OK'
                if not isinstance(interface, str):
                    t = 'error'
                    s = "invalid %s: %s (not a str)" % (key, interface)
                elif len(interface) == 0:
                    t = 'error'
                    s = "'%s' is empty" % key
                self._add_result(t, n, s)
                if t == 'error':
                    continue

            # Check interfaces whitelist.
            t = 'info'
            n = self._get_check_name(iface_type, app=interface, extra=iface)
            s = 'OK'
            if interface not in self.interfaces:
                t = 'error'
                s = "unknown interface '%s'" % interface
            self._add_result(t, n, s)
            if t == 'error':
                continue

            # Abbreviated interfaces don't have attributes, done checking.
            if isinstance(spec, str):
                continue

            # Check interface attributes.
            for attrib in spec:
                if attrib == 'interface':
                    continue
                t = 'info'
                n = self._get_check_name('%s_attributes' % iface_type,
                                         app=iface, extra=attrib)
                s = "OK"
                attrib_key = "%s/%s" % (attrib, iface_type)
                interface_attrs = self.interfaces[interface]
                if attrib_key not in interface_attrs:
                    t = 'error'
                    s = "unknown attribute '%s' for interface '%s' (%s)" % (
                        attrib, interface, iface_type)
                elif not isinstance(spec[attrib],
                                    type(interface_attrs[attrib_key])):
                    t = 'error'
                    s = "'%s' is not '%s'" % \
                        (attrib,
                         type(self.interfaces[interface][attrib_key]).__name__)
                self._add_result(t, n, s)

            # Check required interface attributes
            for req_iface in self.interfaces_required:
                if req_iface != interface or \
                   iface_type not in self.interfaces_required[interface]:
                    continue  # pragma: nocover

                has_required = False
                for req_attrib in \
                        self.interfaces_required[interface][iface_type]:
                    combo = req_attrib.split('/')
                    num_combos = 0     # total number of required attribs
                    found = 0          # required attribs found
                    found_invalid = 0  # must-not-have attribs found

                    for r in combo:
                        if not r.startswith('!'):
                            num_combos += 1

                        if r in spec:
                            found += 1
                        elif r.startswith('!') and r[1:] in spec:
                            found_invalid += 1

                    # if no invalid tokens and we found all the required
                    # attribs then we have a valid combination of required
                    # attribs
                    if found_invalid == 0 and found == num_combos:
                        has_required = True
                        break

                t = 'info'
                n = self._get_check_name('%s_required_attributes' % iface_type,
                                         app=iface)
                if not has_required:
                    # format the combinations for human review
                    combos = []
                    for req_attrib in \
                            self.interfaces_required[interface][iface_type]:
                        combo = []
                        for r in req_attrib.split('/'):
                            if not r.startswith('!'):
                                combo.append(r)
                        combos.append("/".join(combo))
                    t = 'error'
                    s = "missing required %s " % iface_type + \
                        "attributes for interface '%s' " % interface + \
                        "(%s)" % ", ".join(combos)
                self._add_result(t, n, s)

    def check_plugs(self):
        '''Check plugs'''
        iface_type = 'plugs'
        if not self.is_snap2 or iface_type not in self.snap_yaml:
            return

        self._verify_interfaces(iface_type)

    def _verify_app_and_hook_interfaces(self, val, key, hook=False):
        topkey = 'apps'
        topkey_type = 'app'
        if hook:
            topkey = 'hooks'
            topkey_type = 'hook'

        t = 'info'
        n = self._get_check_name("%s_%s" % (topkey_type, key), app=val)
        s = "OK"
        if not isinstance(self.snap_yaml[topkey][val][key], list):
            t = 'error'
            s = "invalid '%s' entry: '%s' (not a list)" % (
                key, self.snap_yaml[topkey][val][key])
        elif len(self.snap_yaml[topkey][val][key]) < 1:
            t = 'error'
            s = "invalid %s entry (empty)" % (key)
        self._add_result(t, n, s)
        if t == 'error':
            return

        # The interface referenced in the entry's 'key' field (plugs/slots) can
        # either be a known interface (when the interface name reference and
        # the interface is the same) or can reference a name in the snap's
        # toplevel 'key' (plugs/slots) mapping
        for ref in self.snap_yaml[topkey][val][key]:
            t = 'info'
            n = self._get_check_name('%s_%s_plug_reference' %
                                     (topkey_type, key),
                                     app=val,
                                     extra=ref)
            s = "OK"
            if not isinstance(ref, str):
                t = 'error'
                s = "invalid %s interface name reference: '%s' (not a str)" \
                    % (key, ref)
            elif ref not in self.interfaces and \
                    (key not in self.snap_yaml or
                     ref not in self.snap_yaml[key]):
                t = 'error'
                s = "unknown %s interface name reference '%s'" % (key, ref)
            self._add_result(t, n, s)
            if t == 'error':
                continue

    def check_apps_plugs(self):
        '''Check apps plugs'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'plugs'
            if key not in self.snap_yaml['apps'][app]:
                continue

            self._verify_app_and_hook_interfaces(app, key)

    def check_hooks_plugs(self):
        '''Check hooks plugs'''
        if not self.is_snap2 or 'hooks' not in self.snap_yaml:
            return

        for hook in self.snap_yaml['hooks']:
            key = 'plugs'
            if self.snap_yaml['hooks'][hook] is None \
                    or key not in self.snap_yaml['hooks'][hook]:
                continue

            self._verify_app_and_hook_interfaces(hook, key, hook=True)

    def check_slots(self):
        '''Check slots'''
        iface_type = 'slots'
        if not self.is_snap2 or iface_type not in self.snap_yaml:
            return

        self._verify_interfaces(iface_type)

    def check_apps_slots(self):
        '''Check apps slots'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'slots'
            if key not in self.snap_yaml['apps'][app]:
                continue

            self._verify_app_and_hook_interfaces(app, key)

    def check_external_symlinks(self):
        '''Check snap for external symlinks'''
        if not self.is_snap2:
            return

        # Note: unclear if gadget snaps can legitimately have external
        # symlinks, but err on side of caution. kernel snaps for reference
        # kernels ship a dangling lib/modules/.../build (like on desktop) but
        # also may have legitimate symlinks in lib/firmware, so just allow
        # them.
        if 'type' in self.snap_yaml and (self.snap_yaml['type'] == 'os' or
                                         self.snap_yaml['type'] == 'kernel'):
            return

        t = 'info'
        n = self._get_check_name('external_symlinks')
        s = 'OK'
        links = find_external_symlinks(self._get_unpack_dir(), self.pkg_files,
                                       self.snap_yaml['name'])
        if len(links) > 0:
            t = 'error'
            s = 'package contains external symlinks: %s' % ', '.join(links)
        self._add_result(t, n, s)

    def check_architecture_all(self):
        '''Check if actually architecture all'''
        if not self.is_snap2:
            return

        if 'architectures' in self.snap_yaml and \
                'all' not in self.snap_yaml['architectures']:
            return

        t = 'info'
        n = self._get_check_name('valid_contents_for_architecture')
        s = 'OK'

        # look for compiled code
        x_binaries = []
        for i in self.pkg_bin_files:
            # .pyc files are arch-independent
            if i.endswith(".pyc"):
                continue
            x_binaries.append(os.path.relpath(i, self._get_unpack_dir()))
        if len(x_binaries) > 0:
            # gadget snap is specified with 'all' but has binaries. Don't
            # complain about that
            t = 'error'
            ok_text = ''
            if 'type' in self.snap_yaml and self.snap_yaml['type'] == 'gadget':
                t = 'info'
                ok_text = " (ok for 'type: gadget')"
            s = "found binaries for architecture 'all': %s%s" % \
                (", ".join(x_binaries), ok_text)
        self._add_result(t, n, s)

    def check_architecture_specified_needed(self):
        '''Check if the specified architecture is actually needed'''
        if not self.is_snap2 or 'architectures' not in self.snap_yaml:
            return

        if 'all' in self.snap_yaml['architectures']:
            return

        for arch in self.snap_yaml['architectures']:
            t = 'info'
            n = self._get_check_name('architecture_specified_needed',
                                     extra=arch)
            s = 'OK'
            if len(self.pkg_bin_files) == 0:
                # This should be a warning but it causes friction for uploads
                t = 'info'
                s = "Could not find compiled binaries for architecture '%s'" \
                    % arch
            self._add_result(t, n, s)

    def check_vcs(self):
        '''Check for VCS files in the package'''
        if not self.is_snap2:
            return

        t = 'info'
        n = self._get_check_name('vcs_files')
        s = 'OK'
        found = []
        for d in self.vcs_files:
            entries = glob.glob("%s/%s" % (self._get_unpack_dir(), d))
            if len(entries) > 0:
                for i in entries:
                    found.append(os.path.relpath(i, self.unpack_dir))
        if len(found) > 0:
            t = 'warn'
            s = 'found VCS files in package: %s' % ", ".join(found)
        self._add_result(t, n, s)

    def check_iffy(self):
        '''Check for iffy files in the package'''
        if not self.is_snap2:
            return

        t = 'info'
        n = self._get_check_name('iffy_files')
        s = 'OK'
        found = []

        for f in self.pkg_files:
            fn = os.path.basename(f)
            for d in self.iffy_files:
                if re.search(r'%s' % d, fn):
                    found.append(os.path.relpath(f, self.unpack_dir))

        if len(found) > 0:
            t = 'warn'
            s = 'found potentially sensitive files in package: %s' % \
                ", ".join(found)
        self._add_result(t, n, s)

    def check_epoch(self):
        '''Check epoch'''
        if not self.is_snap2 or 'epoch' not in self.snap_yaml:
            return

        t = 'info'
        n = self._get_check_name('epoch_valid')
        s = 'OK'
        if not isinstance(self.snap_yaml['epoch'], int):
            t = 'error'
            s = "malformed 'epoch': %s (not an integer)" % (
                self.snap_yaml['epoch'])
        elif int(self.snap_yaml['epoch']) < 0:
            t = 'error'
            s = "malformed 'epoch': '%s' should be positive integer" % (
                self.snap_yaml['epoch'])
        self._add_result(t, n, s)

    def check_confinement(self):
        '''Check confinement'''
        if not self.is_snap2 or 'confinement' not in self.snap_yaml:
            return

        allowed = ['strict', 'devmode', 'classic']
        use_with = ['app', 'gadget', 'kernel']

        t = 'info'
        n = self._get_check_name('confinement_valid')
        s = 'OK'
        manual_review = False
        if not isinstance(self.snap_yaml['confinement'], str):
            t = 'error'
            s = "malformed 'confinement': %s (not a string)" % (
                self.snap_yaml['confinement'])
        elif self.snap_yaml['confinement'] not in allowed:
            t = 'error'
            s = "malformed 'confinement': '%s' should be one of '%s'" % (
                self.snap_yaml['confinement'], ", ".join(allowed))
        elif self.snap_yaml['type'] not in use_with:
            t = 'info'
            s = "'confinement' should not be used with 'type: %s'" % \
                self.snap_yaml['type']
        self._add_result(t, n, s)

        if self.snap_yaml['confinement'] == "classic":
            t = 'info'
            n = self._get_check_name('confinement_classic')
            s = 'OK'
            manual_review = False
            if self.overrides is not None and \
                    'snap_allow_classic' in self.overrides and \
                    self.overrides['snap_allow_classic']:
                s = "OK (confinement '%s' allowed)" % \
                    self.snap_yaml['confinement']
            else:
                t = 'error'
                s = "(NEEDS REVIEW) confinement '%s' not allowed" % \
                    self.snap_yaml['confinement']
                manual_review = True

            self._add_result(t, n, s, manual_review=manual_review)

            t = 'info'
            n = self._get_check_name('confinement_classic_with_interfaces')
            s = 'OK'
            link = None
            found = False
            if 'plugs' in self.snap_yaml or 'slots' in self.snap_yaml:
                found = True
            elif 'apps' in self.snap_yaml:
                for app in self.snap_yaml['apps']:
                    if 'plugs' in self.snap_yaml['apps'][app] or \
                            'slots' in self.snap_yaml['apps'][app]:
                        found = True
                        break
            if found:
                t = 'error'
                s = "confinement '%s' not allowed with plugs/slots" % \
                    self.snap_yaml['confinement']
                link = "https://launchpad.net/bugs/1655369"

            self._add_result(t, n, s, link=link)

    def check_grade(self):
        '''Check confinement'''
        if not self.is_snap2 or 'grade' not in self.snap_yaml:
            return

        allowed = ['stable', 'devel']
# We may reintroduce this, but for now, grade may be used with all snaps
#         use_with = ['app', 'gadget', 'kernel', 'os']

        t = 'info'
        n = self._get_check_name('grade_valid')
        s = 'OK'
        if not isinstance(self.snap_yaml['grade'], str):
            t = 'error'
            s = "malformed 'grade': %s (not a string)" % (
                self.snap_yaml['grade'])
        elif self.snap_yaml['grade'] not in allowed:
            t = 'error'
            s = "malformed 'grade': '%s' should be one of '%s'" % (
                self.snap_yaml['grade'], ", ".join(allowed))
#         elif self.snap_yaml['type'] not in use_with:
#             # "devel" grade os-snap should be avoided.
#             t = 'warn'
#             s = "'grade' should not be used with 'type: %s'" % \
#                 self.snap_yaml['type']
        self._add_result(t, n, s)

    def _verify_env(self, env, app=None):
        t = 'info'
        n = self._get_check_name('environment_valid', app=app)
        s = 'OK'

        if not isinstance(env, dict):
            t = 'error'
            s = "invalid environment: %s (not a dict)" % env
        self._add_result(t, n, s)

        # http://pubs.opengroup.org/onlinepubs/000095399/basedefs/xbd_chap08.html
        invalid = ['=', '\0']
        portable_pat = re.compile(r'^[A-Z_][A-Z0-9_]*$')
        lenient_pat = re.compile(r'^[a-zA-Z0-9_]+$')
        for key in env:
            t = 'info'
            n = self._get_check_name('environment_key_valid', app=app,
                                     extra=key)
            s = 'OK'
            link = None
            invalid_chars = []
            for c in invalid:
                if c in key:
                    invalid_chars.append(c)

            if len(invalid_chars) > 0:
                t = 'error'
                s = "found invalid characters '%s'" % ", ".join(invalid_chars)
            elif not portable_pat.search(key) and lenient_pat.search(key):
                t = 'info'
                s = "'%s' is not shell portable" % key
                link = ("http://pubs.opengroup.org/onlinepubs/000095399/"
                        "basedefs/xbd_chap08.html")
            elif not lenient_pat.search(key):
                t = 'warn'
                s = "unusual characters in '%s' " % key + \
                    "(should be '^[a-zA-Z0-9_]+$')"
            self._add_result(t, n, s, link=link)

            # The only limit on the contents of an arg appear to be the length
            # but that is going to be language and system dependent, so don't
            # worry about it here (this would simply be a bug in the software)
            t = 'info'
            n = self._get_check_name('environment_value_valid', app=app,
                                     extra=key)
            s = 'OK'
            if not isinstance(env[key], str) and \
                    not isinstance(env[key], int) and \
                    not isinstance(env[key], float):
                t = 'error'
                s = "invalid environment value for '%s': %s" % (key, env[key])
            self._add_result(t, n, s)

    def check_environment(self):
        '''Check environment'''
        if not self.is_snap2 or 'environment' not in self.snap_yaml:
            return

        self._verify_env(self.snap_yaml['environment'])

    def check_apps_environment(self):
        '''Check apps environment'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'environment'
            if key not in self.snap_yaml['apps'][app]:
                continue

            self._verify_env(self.snap_yaml['apps'][app]['environment'],
                             app=app)

    def check_apps_aliases(self):
        '''Check apps aliases'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        seen = []
        for app in self.snap_yaml['apps']:
            key = 'aliases'
            if key not in self.snap_yaml['apps'][app]:
                continue

            aliases = self.snap_yaml['apps'][app]['aliases']

            t = 'info'
            n = self._get_check_name('aliases_valid', app=app)
            s = 'OK'
            if not isinstance(aliases, list):
                t = 'error'
                s = "invalid aliases: %s (not a list)" % aliases
            elif len(aliases) == 0:
                t = 'error'
                s = 'invalid aliases (empty)'
            self._add_result(t, n, s)

            # from validate.go in snapd
            pat = re.compile(r'^[a-zA-Z0-9][-_.a-zA-Z0-9]*$')
            for alias in aliases:
                t = 'info'
                n = self._get_check_name('alias_valid', app=app,
                                         extra=alias)
                if not pat.search(alias):
                    t = 'error'
                    s = "malformed alias '%s' " % alias + \
                        "(should be '^[a-zA-Z0-9][-_.a-zA-Z0-9]*$')"
                elif alias in seen:
                    t = 'error'
                    s = "alias '%s' used more than once" % alias
                self._add_result(t, n, s)
                seen.append(alias)

    def _uses_interface(self, iface_type, iface):
        '''Get interface name by type and interface/interface reference.
           Returns:
           - The dereferenced interface name
           - None if could not be found
        '''
        if iface_type in self.snap_yaml:
            for ref in self.snap_yaml[iface_type]:
                interface = ''
                spec = self.snap_yaml[iface_type][ref]
                if isinstance(spec, str):
                    # Abbreviated syntax (no attributes)
                    # <plugs|slots>:
                    #   <alias>: <interface>
                    interface = spec
                elif 'interface' in spec:
                    # Full specification.
                    # <plugs|slots>:
                    #   <alias>:
                    #     interface: <interface>
                    interface = spec['interface']
                elif isinstance(spec, dict):
                    # Abbreviated syntax (no attributes)
                    # <plugs|slots>:
                    #   <interface>: null
                    if len(spec) == 0:
                        interface = ref
                if interface == iface:
                    return True

        if 'apps' in self.snap_yaml:
            for app in self.snap_yaml['apps']:
                if iface_type in self.snap_yaml['apps'][app] and \
                        iface in self.snap_yaml['apps'][app][iface_type]:
                    return True

        return False

    def _verify_desktop_file(self, fn):
        '''Verify the desktop file'''
        if 'apps' not in self.snap_yaml:
            return

        appnames = []
        for app in self.snap_yaml['apps']:
            if app == self.snap_yaml['name']:
                appnames.append(app)
            else:
                appnames.append("%s.%s" % (self.snap_yaml['name'], app))

        fh = self._extract_file(fn)

        # For now, just check Exec= since snapd strips out anything it
        # doesn't understand. TODO: implement full checks
        found_exec = False
        t = 'info'
        n = self._get_check_name('desktop_file',
                                 extra=os.path.basename(fn))
        s = 'OK'
        for line in fh.readlines():
            line = line.rstrip()
            if line.startswith('Exec='):
                found_exec = True
                break
        if not found_exec:
            t = 'error'
            s = "Could not find 'Exec=' in desktop file"
        self._add_result(t, n, s)

    def check_meta_gui_desktop(self):
        '''Check meta/gui/*.desktop'''
        if not self.is_snap2:
            return

        has_desktop_files = False
        for f in self.pkg_files:
            fn = os.path.relpath(f, self._get_unpack_dir())
            if fn.startswith("meta/gui/") and fn.endswith(".desktop"):
                self._verify_desktop_file(f)
                has_desktop_files = True
                break

        desktop_interfaces_specified = []
        for iface in self.interface_plug_requires_desktop_file:
            if self._uses_interface("plugs", iface):
                desktop_interfaces_specified.append(iface)

        if len(desktop_interfaces_specified) < 1 and not has_desktop_files:
            return

        t = 'info'
        n = self._get_check_name('meta_gui_desktop')
        s = 'OK'
        if len(desktop_interfaces_specified) > 0 and not has_desktop_files:
            if self.snap_yaml['name'] in desktop_file_exception:
                t = 'info'
                s = "OK (overidden)"
            else:
                t = 'warn'
                s = ("desktop interfaces "
                     "(%s) " % ",".join(desktop_interfaces_specified) +
                     "specified without a corresponding meta/gui/*.desktop "
                     "file. If using snapcraft, please see "
                     "https://snapcraft.io/docs/build-snaps/metadata"
                     "#fixed-assets. "
                     "Otherwise, please provide a desktop file in "
                     "meta/gui/*.desktop (it should reference one of the "
                     "'apps' from your snapcraft/snap.yaml).")

        self._add_result(t, n, s)

    def check_apps_completer(self):
        '''Check apps - completer'''
        if not self.is_snap2 or 'apps' not in self.snap_yaml:
            return

        for app in self.snap_yaml['apps']:
            key = 'completer'
            if key not in self.snap_yaml['apps'][app]:
                continue

            self._verify_value_is_file(app, key)



'''sr_tests.py: common setup and tests for test modules'''
#
# Copyright (C) 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import os
import yaml

from unittest.mock import patch
from unittest import TestCase
from clickreviews.common import (
    check_results as common_check_results
)

# These should be set in the test cases
TEST_SNAP_YAML = ""
TEST_PKGFMT_TYPE = "snap"
TEST_PKGFMT_VERSION = "16.04"
TEST_UNPACK_DIR = "/fake"
TEST_SECURITY_PROFILES = dict()
TEST_UNSQUASHFS_LLS = ""


#
# Mock override functions
#
def _mock_func(self):
    '''Fake test function'''
    return


def _extract_snap_yaml(self):
    '''Pretend we read the package.yaml file'''
    return io.StringIO(TEST_SNAP_YAML)


def _path_join(self, d, fn):
    '''Pretend we have a tempdir'''
    return os.path.join("/fake", fn)


def _check_innerpath_executable(self, fn):
    '''Pretend we a file'''
    if '.nonexec' in fn:
        return False
    return True


def _pkgfmt_type(self):
    '''Pretend we found the pkgfmt type'''
    return TEST_PKGFMT_TYPE


def _detect_package(self, fn):
    '''Pretend we detected the package'''
    ver = 2
    if TEST_PKGFMT_VERSION == "15.04":
        ver = 1
    return (TEST_PKGFMT_TYPE, ver)


def __get_unpack_dir(self):
    '''Pretend we found the unpack dir'''
    return TEST_UNPACK_DIR


def _unsquashfs_lls(self, fn):
    '''Pretend we ran unsquashfs -lls fn'''
    return (0, TEST_UNSQUASHFS_LLS)


def create_patches():
    # http://docs.python.org/3.4/library/unittest.mock-examples.html
    # Mock patching. Don't use decorators but instead patch in setUp() of the
    # child.
    patches = []
    patches.append(patch('clickreviews.common.Review._check_package_exists',
                   _mock_func))
    patches.append(patch(
        'clickreviews.sr_common.SnapReview._extract_snap_yaml',
        _extract_snap_yaml))
    patches.append(patch(
        'clickreviews.sr_common.SnapReview._path_join',
        _path_join))
    patches.append(patch('clickreviews.common.unpack_pkg', _mock_func))
    patches.append(patch('clickreviews.common.raw_unpack_pkg', _mock_func))
    patches.append(patch('clickreviews.common.detect_package',
                   _detect_package))
    patches.append(patch('clickreviews.sr_common.SnapReview._list_all_files',
                   _mock_func))
    patches.append(patch(
        'clickreviews.sr_common.SnapReview._list_all_compiled_binaries',
        _mock_func))

    patches.append(patch('clickreviews.common.Review._list_all_files',
                   _mock_func))
    patches.append(patch(
        'clickreviews.common.Review._list_all_compiled_binaries',
        _mock_func))
    patches.append(patch(
        'clickreviews.common.Review._check_innerpath_executable',
        _check_innerpath_executable))

    # sr_common
    patches.append(patch('clickreviews.sr_common.SnapReview._get_unpack_dir',
                   __get_unpack_dir))
    patches.append(patch("clickreviews.sr_common.SnapReview._pkgfmt_type",
                   _pkgfmt_type))

    # sr_security
    patches.append(patch(
        "clickreviews.sr_security.SnapReviewSecurity._unsquashfs_lls",
        _unsquashfs_lls))

    return patches


class TestSnapReview(TestCase):
    """Tests for the snap review tool."""
    def __init__(self, *args):
        TestCase.__init__(self, *args)
        self._reset_test_data()

    def _reset_test_data(self):
        self.test_snap_yaml = dict()
        self.set_test_pkgfmt("snap", "16.04")

        self.set_test_snap_yaml("name", "foo")
        self.set_test_snap_yaml("version", "0.1")
        self.set_test_snap_yaml("description", "Test description")
        self.set_test_snap_yaml("summary", "Test summary")
        self.set_test_snap_yaml("architectures", ["all"])
        apps = dict()
        apps["bar"] = dict()
        apps["bar"]["command"] = "bin/bar"
        self.set_test_snap_yaml("apps", apps)

        # mockup a package name
        self._update_test_name()

        # reset the security profiles
        self.test_security_profiles = dict()

    def _update_test_snap_yaml(self):
        global TEST_SNAP_YAML
        TEST_SNAP_YAML = yaml.dump(self.test_snap_yaml,
                                   default_flow_style=False,
                                   indent=4)

    def _update_test_security_profiles(self):
        global TEST_SECURITY_PROFILES
        TEST_SECURITY_PROFILES = dict()
        if len(self.test_security_profiles.keys()) == 0:
            TEST_SECURITY_PROFILES = dict()
        else:
            for plug in self.test_security_profiles.keys():
                TEST_SECURITY_PROFILES[plug] = \
                    self.test_security_profiles[plug]

    def _update_test_name(self):
        self.test_name = "%s.origin_%s_%s.snap" % (
            self.test_snap_yaml["name"],
            self.test_snap_yaml["version"],
            self.test_snap_yaml["architectures"][0])

    def check_results(self, report,
                      expected_counts={'info': 1, 'warn': 0, 'error': 0},
                      expected=None):
        common_check_results(self, report, expected_counts, expected)

    def check_manual_review(self, report, check_name,
                            result_type='error', manual_review=True):
        result = report[result_type][check_name]
        self.assertEqual(result['manual_review'], manual_review)

    def set_test_snap_yaml(self, key, value):
        '''Set key in meta/snap.yaml to value. If value is None, remove
           key'''
        if value is None:
            if key in self.test_snap_yaml:
                self.test_snap_yaml.pop(key, None)
        else:
            self.test_snap_yaml[key] = value
        self._update_test_snap_yaml()

    def set_test_security_profile(self, plug, key, policy):
        '''Set policy in security profile for key'''
        if policy is None:
            if plug in self.test_security_profiles and \
                    key in self.test_security_profiles[plug]:
                self.test_security_profiles[plug].pop(key)
        elif key is None:
            if plug in self.test_security_profiles:
                self.test_security_profiles.pop(plug)
        elif plug is None:
            self.test_security_profiles = dict()
        else:
            if plug not in self.test_security_profiles:
                self.test_security_profiles[plug] = dict()
            self.test_security_profiles[plug][key] = policy

        self._update_test_security_profiles()

    def set_test_pkgfmt(self, t, v):
        global TEST_PKGFMT_TYPE
        global TEST_PKGFMT_VERSION
        TEST_PKGFMT_TYPE = t
        TEST_PKGFMT_VERSION = v

    def set_test_unpack_dir(self, d):
        global TEST_UNPACK_DIR
        TEST_UNPACK_DIR = d

    def set_test_unsquashfs_lls(self, s):
        global TEST_UNSQUASHFS_LLS
        TEST_UNSQUASHFS_LLS = s

    def setUp(self):
        '''Make sure our patches are applied everywhere'''
        patches = create_patches()
        for p in patches:
            p.start()
            self.addCleanup(p.stop)

    def tearDown(self):
        '''Make sure we reset everything to known good values'''
        global TEST_SNAP_YAML
        TEST_SNAP_YAML = ""
        global TEST_SECURITY_PROFILES
        TEST_SECURITY_PROFILES = dict()
        global TEST_PKGFMT_TYPE
        TEST_PKGFMT_TYPE = "snap"
        global TEST_PKGFMT_VERSION
        TEST_PKGFMT_VERSION = "16.04"
        global TEST_UNPACK_DIR
        TEST_UNPACK_DIR = "/fake"
        global TEST_UNSQUASHFS_LLS
        TEST_UNSQUASHFS_LLS = ""

        self._reset_test_data()
